# coding=utf-8
#
# Copyright (C) 2018 Martin Owens <doctormo@gmail.com>
# Copyright (C) 2023 Jonathan Neuhauser <jonathan.neuhauser@outlook.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""Curve and Smooth Path Commands"""

from __future__ import annotations

from typing import overload, Tuple, Callable, Union, cast

import numpy as np

from ..transforms import cubic_extrema, Transform, Vector2d, ComplexLike

from .interfaces import (
    AbsolutePathCommand,
    RelativePathCommand,
    BezierArcComputationMixin,
    BezierComputationMixin,
)


class CurveMixin(BezierComputationMixin, BezierArcComputationMixin):
    """Common functionality for curves"""

    ccontrol_points: Callable[[complex, complex, complex], Tuple[complex, ...]]

    def ccurve_points(
        self, first: complex, prev: complex, prev_prev: complex
    ) -> Tuple[complex, ...]:
        """Common implementation of ccurve_points for Curves"""
        return self.ccontrol_points(first, prev, prev_prev)

    def _cderivative(
        self, first: complex, prev: complex, prev_control: complex, t: float, n: int = 1
    ) -> complex:
        """Returns the nth derivative of the segment at t.

        .. hint:: Bezier curves can have points where their derivative vanishes.
            If you are interested in the tangent direction, use the :func:`unit_tangent`
            method instead.
        """

        points = self.ccontrol_points(first, prev, prev_control)

        if n == 1:
            return (
                3 * (points[0] - prev) * ((1 - t) ** 2)
                + 6 * (points[1] - points[0]) * (1 - t) * t
                + 3 * (points[2] - points[1]) * t**2
            )
        elif n == 2:
            return 6 * (
                (1 - t) * (points[1] - 2 * points[0] + prev)
                + t * (points[2] - 2 * points[1] + points[0])
            )
        elif n == 3:
            return 6 * (points[2] - 3 * (points[1] - points[0]) - prev)
        elif n > 3:
            return complex(0, 0)
        else:
            raise ValueError("n should be a positive integer.")

    def poly(self, prev, prev_control, return_coeffs=False):
        """Returns a the cubic as a complex Polynomial object.

        .. versionadded:: 1.4
        """
        points = self.ccontrol_points(0j, prev, prev_control)
        coeffs = (
            -prev + 3 * (points[0] - points[1]) + points[2],
            3 * (prev - 2 * points[0] + points[1]),
            3 * (prev + points[0]),
            prev,
        )
        if return_coeffs:
            return coeffs
        return np.poly1d(coeffs)

    def _cunit_tangent(
        self, first: complex, prev: complex, prev_control: complex, t: float
    ) -> complex:
        return self.bezier_unit_tangent(prev, prev_control, t)

    def _curvature(
        self, first: complex, prev: complex, prev_control: complex, t: float
    ):
        return self.segment_curvature(prev, prev_control, t)

    def _abssplit(
        self, prev: complex, prev_control: complex, t: float
    ) -> Tuple[Curve, Curve]:
        """Split this curve and return two Curves using De Casteljau's algorithm"""
        p1, p2, p3 = self.ccontrol_points(0j, prev, prev_control)
        p1_1 = (1 - t) * prev + t * p1
        p1_2 = (1 - t) * p1 + t * p2
        p1_3 = (1 - t) * p2 + t * p3
        p2_1 = (1 - t) * p1_1 + t * p1_2
        p2_2 = (1 - t) * p1_2 + t * p1_3
        p3_1 = (1 - t) * p2_1 + t * p2_2

        return Curve(p1_1, p2_1, p3_1), Curve(p2_2, p1_3, p3)

    def _relsplit(self, prev: complex, prev_control: complex, t: float):
        """Split this curve and return two curves"""
        c1abs, c2abs = self._abssplit(prev, prev_control, t)
        return c1abs.to_relative(prev), c2abs.to_relative(c1abs.arg3)

    def _cpoint(
        self, first: complex, prev: complex, prev_control: complex, t: float
    ) -> complex:
        control1, control2, end = self.ccontrol_points(first, prev, prev_control)
        return (
            (1 - t) ** 3 * prev
            + 3 * t * (1 - t) ** 2 * control1
            + 3 * t**2 * (1 - t) * control2
            + t**3 * end
        )


class Curve(CurveMixin, AbsolutePathCommand):
    """Absolute Curved Line segment"""

    letter = "C"
    nargs = 6

    arg1: complex
    """The (absolute) first control point"""

    arg2: complex
    """The (absolute) second control point"""

    arg3: complex
    """The (absolute) end point"""

    @property
    def x2(self) -> float:
        """x coordinate of the (absolute) first control point"""
        return self.arg1.real

    @property
    def y2(self) -> float:
        """y coordinate of the (absolute) first control point"""
        return self.arg1.imag

    @property
    def x3(self) -> float:
        """x coordinate of the (absolute) second control point"""
        return self.arg2.real

    @property
    def y3(self) -> float:
        """y coordinate of the (absolute) second control point"""
        return self.arg2.imag

    @property
    def x4(self) -> float:
        """x coordinate of the (absolute) end point"""
        return self.arg3.real

    @property
    def y4(self) -> float:
        """y coordinate of the (absolute) end point"""
        return self.arg3.imag

    @property
    def args(self):
        return (
            self.arg1.real,
            self.arg1.imag,
            self.arg2.real,
            self.arg2.imag,
            self.arg3.real,
            self.arg3.imag,
        )

    @overload
    def __init__(self, x2: ComplexLike, x3: ComplexLike, x4: ComplexLike): ...

    @overload
    def __init__(
        self, x2: float, y2: float, x3: float, y3: float, x4: float, y4: float
    ): ...  # pylint: disable=too-many-arguments

    def __init__(self, x2, y2, x3, y3=None, x4=None, y4=None):  # pylint: disable=too-many-arguments
        if y3 is not None:
            self.arg1 = x2 + y2 * 1j
            self.arg2 = x3 + y3 * 1j
            self.arg3 = x4 + y4 * 1j
        else:
            self.arg1, self.arg2, self.arg3 = complex(x2), complex(y2), complex(x3)

    def update_bounding_box(self, first, last_two_points, bbox):
        x1, x2, x3, x4 = last_two_points[-1].real, self.x2, self.x3, self.x4
        y1, y2, y3, y4 = last_two_points[-1].imag, self.y2, self.y3, self.y4

        if not (x1 in bbox.x and x2 in bbox.x and x3 in bbox.x and x4 in bbox.x):
            bbox.x += cubic_extrema(x1, x2, x3, x4)

        if not (y1 in bbox.y and y2 in bbox.y and y3 in bbox.y and y4 in bbox.y):
            bbox.y += cubic_extrema(y1, y2, y3, y4)

    def transform(self, transform: Transform) -> Curve:
        return Curve(
            transform.capply_to_point(self.arg1),
            transform.capply_to_point(self.arg2),
            transform.capply_to_point(self.arg3),
        )

    def ccontrol_points(
        self, first: complex, prev: complex, prev_prev: complex
    ) -> Tuple[complex, ...]:
        # pylint: disable=unused-argument
        return (self.arg1, self.arg2, self.arg3)

    def to_relative(self, prev: ComplexLike) -> curve:
        return curve(self.arg1 - prev, self.arg2 - prev, self.arg3 - prev)

    def cend_point(self, first: complex, prev: complex) -> complex:
        return self.arg3

    def reverse(self, first: ComplexLike, prev: ComplexLike) -> Curve:
        return Curve(self.arg2, self.arg1, prev)

    def to_bez(self):
        """Convert to [[c1x, c1y], [c2x, c2y], [end_x, end_y]]"""
        return [Vector2d.c2t(i) for i in self.ccontrol_points(0j, 0j, 0j)]

    def _split(
        self, first: complex, prev: complex, prev_control: complex, t: float
    ) -> Tuple[Curve, Curve]:
        return self._abssplit(prev, prev_control, t)


class curve(CurveMixin, RelativePathCommand):  # pylint: disable=invalid-name
    """Relative curved line segment"""

    letter = "c"
    nargs = 6

    arg1: complex
    """The (relative) first control point"""

    arg2: complex
    """The (relative) second control point"""

    arg3: complex
    """The (relative) end point"""

    @property
    def dx2(self) -> float:
        """x coordinate of the (relative) first control point"""
        return self.arg1.real

    @property
    def dy2(self) -> float:
        """y coordinate of the (relative) first control point"""
        return self.arg1.imag

    @property
    def dx3(self) -> float:
        """x coordinate of the (relative) second control point"""
        return self.arg2.real

    @property
    def dy3(self) -> float:
        """y coordinate of the (relative) second control point"""
        return self.arg2.imag

    @property
    def dx4(self) -> float:
        """x coordinate of the (relative) end point"""
        return self.arg3.real

    @property
    def dy4(self) -> float:
        """y coordinate of the (relative) end point"""
        return self.arg3.imag

    @overload
    def __init__(self, dx2: ComplexLike, dx3: ComplexLike, dx4: ComplexLike): ...

    @overload
    def __init__(
        self, dx2: float, dy2: float, dx3: float, dy3: float, dx4: float, dy4: float
    ): ...  # pylint: disable=too-many-arguments

    def __init__(self, dx2, dy2, dx3, dy3=None, dx4=None, dy4=None):  # pylint: disable=too-many-arguments
        if dy3 is not None:
            self.arg1 = dx2 + dy2 * 1j
            self.arg2 = dx3 + dy3 * 1j
            self.arg3 = dx4 + dy4 * 1j
        else:
            self.arg1, self.arg2, self.arg3 = complex(dx2), complex(dy2), complex(dx3)

    @property
    def args(self):
        return self.dx2, self.dy2, self.dx3, self.dy3, self.dx4, self.dy4

    def to_absolute(self, prev: ComplexLike) -> Curve:
        return Curve(*self.ccurve_points(0j, complex(prev), 0j))

    def cend_point(self, first: complex, prev: complex) -> complex:
        return self.arg3 + prev

    def reverse(self, first: ComplexLike, prev: ComplexLike) -> curve:
        return curve(-self.arg3 + self.arg2, -self.arg3 + self.arg1, -self.arg3)

    def ccontrol_points(
        self, first: complex, prev: complex, prev_prev: complex
    ) -> Tuple[complex, ...]:
        # pylint: disable=unused-argument
        return (
            self.arg1 + prev,
            self.arg2 + prev,
            self.arg3 + prev,
        )

    def _split(
        self, first: complex, prev: complex, prev_control: complex, t: float
    ) -> Tuple[curve, curve]:
        return self._relsplit(prev, prev_control, t)


class Smooth(CurveMixin, AbsolutePathCommand):
    """Absolute Smoothed Curved Line segment"""

    letter = "S"
    nargs = 4

    arg1: complex
    """The (absolute) control point"""

    arg2: complex
    """The (absolute) end point"""

    @property
    def x3(self) -> float:
        """x coordinate of the (absolute) control point"""
        return self.arg1.real

    @property
    def y3(self) -> float:
        """y coordinate of the (absolute) control point"""
        return self.arg1.imag

    @property
    def x4(self) -> float:
        """x coordinate of the (absolute) end point"""
        return self.arg2.real

    @property
    def y4(self) -> float:
        """y coordinate of the (absolute) end point"""
        return self.arg2.imag

    @property
    def args(self):
        return self.x3, self.y3, self.x4, self.y4

    @overload
    def __init__(self, x3: ComplexLike, x4: ComplexLike): ...

    @overload
    def __init__(self, x3: float, y3: float, x4: float, y4: float): ...

    def __init__(self, x3, y3, x4=None, y4=None):
        if x4 is not None:
            self.arg1 = x3 + y3 * 1j
            self.arg2 = x4 + y4 * 1j
        else:
            self.arg1, self.arg2 = complex(x3), complex(y3)

    def update_bounding_box(self, first, last_two_points, bbox):
        # pylint: disable=no-member
        self.to_curve(last_two_points[-1], last_two_points[-2]).update_bounding_box(
            first, last_two_points, bbox
        )

    def ccontrol_points(
        self, first: complex, prev: complex, prev_prev: complex
    ) -> Tuple[complex, ...]:
        # pylint: disable=unused-argument
        return (2 * prev - prev_prev, self.arg1, self.arg2)

    def to_non_shorthand(self, prev: ComplexLike, prev_control: ComplexLike) -> Curve:
        return self.to_curve(prev, prev_control)

    def to_relative(self, prev: ComplexLike) -> smooth:
        return smooth(self.arg1 - prev, self.arg2 - prev)

    def transform(self, transform: Transform) -> Smooth:
        return Smooth(
            transform.capply_to_point(self.arg1), transform.capply_to_point(self.arg2)
        )

    def cend_point(self, first: complex, prev: complex) -> complex:
        return self.arg2

    def reverse(self, first: ComplexLike, prev: ComplexLike) -> Smooth:
        return Smooth(self.arg1, prev)

    def _split(
        self, first: complex, prev: complex, prev_control: complex, t: float
    ) -> Tuple[Curve, Curve]:
        # We can't preserve the smooth type for a split because de Casteljau's
        # algorithm changes the handles (obviously).
        # Only in special cases such as splitting to subsequent smooth segments at
        # t=1/2 such a preservation would be possible
        crv = cast(Curve, self.to_non_shorthand(prev, prev_control))
        return crv._split(  # pylint: disable=protected-access, no-member
            first, prev, prev_control, t
        )


class smooth(CurveMixin, RelativePathCommand):  # pylint: disable=invalid-name
    """Relative smoothed curved line segment"""

    letter = "s"
    nargs = 4

    arg1: complex
    """The (absolute) control point"""

    arg2: complex
    """The (absolute) end point"""

    @property
    def dx3(self) -> float:
        """x coordinate of the (relative) control point"""
        return self.arg1.real

    @property
    def dy3(self) -> float:
        """y coordinate of the (relative) control point"""
        return self.arg1.imag

    @property
    def dx4(self) -> float:
        """x coordinate of the (relative) end point"""
        return self.arg2.real

    @property
    def dy4(self) -> float:
        """y coordinate of the (relative) end point"""
        return self.arg2.imag

    @property
    def args(self):
        return self.dx3, self.dy3, self.dx4, self.dy4

    @overload
    def __init__(self, dx3: ComplexLike, dx4: ComplexLike): ...

    @overload
    def __init__(self, dx3: float, dy3: float, dx4: float, dy4: float): ...

    def __init__(self, dx3, dy3, dx4=None, dy4=None):
        if dx4 is not None:
            self.arg1 = dx3 + dy3 * 1j
            self.arg2 = dx4 + dy4 * 1j
        else:
            self.arg1, self.arg2 = complex(dx3), complex(dy3)

    def to_absolute(self, prev: ComplexLike) -> Smooth:
        return Smooth(self.arg1 + prev, self.arg2 + prev)

    def cend_point(self, first: complex, prev: complex) -> complex:
        return self.arg2 + prev

    def to_non_shorthand(self, prev: ComplexLike, prev_control: ComplexLike) -> Curve:
        return self.to_absolute(prev).to_non_shorthand(prev, prev_control)

    def reverse(self, first: ComplexLike, prev: ComplexLike):
        return smooth(-self.arg2 + self.arg1, -self.arg2)

    def ccontrol_points(
        self, first: complex, prev: complex, prev_prev: complex
    ) -> Tuple[complex, ...]:
        # pylint: disable=unused-argument
        return (2 * prev - prev_prev, self.arg1 + prev, self.arg2 + prev)

    def _split(
        self, first: complex, prev: complex, prev_control: complex, t: float
    ) -> Tuple[curve, curve]:
        return self._relsplit(prev, prev_control, t)



#
# Copyright 2012-2022 Martin Owens <doctormo@geek-2.com>
#
# This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>
#
# pylint: disable=too-many-instance-attributes
"""
Wraps the gtk windows with something a little nicer.
"""

import logging

from gi.repository import Gtk

PROPS = {
    "Box": ["expand", "fill", "padding", "pack-type"],
    "Grid": ["top-attach", "left-attach", "height", "width"],
    "Table": ["top-attach", "left-attach", "bottom-attach", "right-attach"],
}


def protect(cls, *methods):
    """Simple check for protecting an inherited class from having
    certain methods over-ridden"""
    if not isinstance(cls, type):
        cls = type(cls)
    for method in methods:
        if method in cls.__dict__:  # pragma: no cover
            raise RuntimeError(
                f"{cls.__name__} in {cls.__module__} has protected def {method}()"
            )


class Window:
    """
    This wraps gtk windows and allows for having parent windows.
    """

    primary = True
    """Should the window be the first loaded and end gtk when it is closed."""

    name = None
    """The name of the window"""

    def __init__(self, gapp):
        self.gapp = gapp
        self.dead = False
        self.parent = None
        self.args = ()
        ui_file = gapp.get_ui_file(self.name)

        # Setup the gtk app connection
        self.w_tree = Gtk.Builder()
        self.widget = self.w_tree.get_object
        self.w_tree.set_translation_domain(gapp.app_name)
        self.w_tree.add_from_file(ui_file)

        # Setup the gtk builder window
        self.window = self.widget(self.name)
        if not self.window:  # pragma: no cover
            raise KeyError(f"Missing window widget '{self.name}' from '{ui_file}'")

        # Give us a window id to track this window
        self.wid = str(hash(self.window))

    def extract(self):
        """Extract this window's container for use in other apps"""
        for child in self.window.get_children():
            self.window.remove(child)
            return child

    def init(self, parent=None, **kwargs):
        """Initialise the window within the GtkApp"""
        if "replace" not in kwargs:
            protect(self, "destroy", "exit", "load_window", "proto_window")
        self.args = kwargs
        # Set object defaults
        self.parent = parent

        self.w_tree.connect_signals(self)

        # These are some generic convience signals
        self.window.connect("destroy", self.exit)

        # If we have a parent window, then we expect not to quit
        if self.parent:
            self.window.set_transient_for(self.parent)
            self.parent.set_sensitive(False)

        # We may have some more gtk widgets to setup
        self.load_widgets(**self.args)
        self.window.show()

    def load_window(self, name, *args, **kwargs):
        """Load child window, automatically sets parent"""
        kwargs["parent"] = self.window
        return self.gapp.load_window(name, *args, **kwargs)

    def load_widgets(self):
        """Child class should use this to create widgets"""

    def destroy(self, widget=None):  # pylint: disable=unused-argument
        """Destroy the window"""
        logging.debug("Destroying Window '%s'", self.name)
        self.window.destroy()
        # We don't need to call self.exit(), handeled by window event.

    def pre_exit(self):
        """Internal method for what to do when the window has died"""

    def exit(self, widget=None):
        """Called when the window needs to exit."""
        # Is the signal called by the window or by something else?
        if not widget or not isinstance(widget, Gtk.Window):
            self.destroy()
        # Clean up any required processes
        self.pre_exit()
        if self.parent:
            # We assume the parent didn't load another gtk loop
            self.parent.set_sensitive(True)
        # Exit our entire app if this is the primary window
        # Or just remove from parent window list, which may still exit.
        if self.primary:
            logging.debug("Exiting the application")
            self.gapp.exit()
        else:
            logging.debug("Removing Window %s from parent", self.name)
            self.gapp.remove_window(self)
        # Now finish up what ever is left to do now the window is dead.
        self.dead = True
        self.post_exit()
        return widget

    def post_exit(self):
        """Called after we've killed the window"""

    def if_widget(self, name):
        """
        Attempt to get the widget from gtk, but if not return a fake that won't
        cause any trouble if we don't further check if it's real.
        """
        return self.widget(name) or FakeWidget(name)

    def replace(self, old, new):
        """Replace the old widget with the new widget"""
        if isinstance(old, str):
            old = self.widget(old)
        if isinstance(new, str):
            new = self.widget(new)
        target = old.get_parent()
        source = new.get_parent()
        if target is not None:
            if source is not None:
                source.remove(new)
            target.remove(old)
            target.add(new)

    @staticmethod
    def get_widget_name(obj):
        """Return the widget's name in the builder file"""
        return Gtk.Buildable.get_name(obj)


class ChildWindow(Window):
    """Base class for child window objects.

    These child windows are typically
    window objects in the same gtk builder file as their parents. If you just want
    to make a window that interacts with a parent window, use the normal
    Window class and call with the optional parent attribute.
    """

    primary = False


class FakeWidget:
    """A fake widget class that can take calls"""

    def __init__(self, name):
        self._name = name

    def __getattr__(self, name):
        def _fake(*args, **kwargs):
            logging.info("Calling fake method: %s:%s", args, kwargs)

        return _fake

    def __bool__(self):
        return False



# coding=utf-8
#
# Copyright (C) 2018 Martin Owens <doctormo@gmail.com>
# Copyright (C) 2023 Jonathan Neuhauser <jonathan.neuhauser@outlook.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""Interfaces for path commands"""

from __future__ import annotations

import abc
import math
from typing import (
    List,
    Any,
    Tuple,
    Dict,
    Type,
    Generator,
    Union,
    TYPE_CHECKING,
    Optional,
    Callable,
)
from dataclasses import dataclass

import numpy as np
from ..utils import classproperty, rational_limit
from ..transforms import Vector2d, BoundingBox, Transform, ComplexLike

if TYPE_CHECKING:
    from .curves import Curve
    from .lines import Line


@dataclass
class LengthSettings:
    """Settings for :func:`PathCommand.length`

    .. versionadded::  1.4
    """

    min_depth: int = 5
    error: float = 1e-5


@dataclass
class ILengthSettings:
    """Settings for :func:`PathCommand.ilength`

    .. versionadded:: 1.4
    """

    min_depth: int = 5

    error: float = 1e-5
    """
    Error tolerance for the computations of the test segment that is performed
    for each iteration.

    The defaults from svgpathtools are ILENGTH_ERROR=ILENGTH_LENGTH_TOL=1e-12.
    This is rather slow, particularly _length (which then subdivides the path into
    2^12 or more segments and adds up the length). For visual editing, this is rather
    irrelevant (and a lot more accurate than the previous methods)."""

    length_tol: float = 1e-5
    """Total (absolute) tolerance of the resulting s value."""

    maxits: int = 10000


class PathCommand(abc.ABC):
    """
    Base class of all path commands
    """

    letter = ""
    # Number of arguments that follow this path commands letter
    nargs = -1

    @classproperty  # From python 3.9 on, just combine @classmethod and @property
    def name(cls):  # pylint: disable=no-self-argument
        """The full name of the segment (i.e. Line, Arc, etc)"""
        return cls.__name__  # pylint: disable=no-member

    @classproperty
    def next_command(self):
        """The implicit next command. This is for automatic chains where the next
        command isn't given, just a bunch on numbers which we automatically parse."""
        return self

    @property
    def is_relative(self) -> bool:
        """Whether the command is defined in relative coordinates, i.e. relative to
        the previous endpoint (lower case path command letter)"""
        raise NotImplementedError

    @property
    def is_absolute(self) -> bool:
        """Whether the command is defined in absolute coordinates (upper case path
        command letter)"""
        raise NotImplementedError

    def to_relative(self, prev: ComplexLike) -> RelativePathCommand:
        """Return absolute counterpart for absolute commands or copy for relative"""
        raise NotImplementedError

    def to_absolute(self, prev: ComplexLike) -> AbsolutePathCommand:
        """Return relative counterpart for relative commands or copy for absolute"""
        raise NotImplementedError

    def reverse(self, first: ComplexLike, prev: ComplexLike) -> PathCommand:
        """Reverse path command

        .. versionadded:: 1.1
        """
        raise NotImplementedError

    def to_non_shorthand(
        self,
        prev: ComplexLike,
        prev_control: ComplexLike,  # pylint: disable=unused-argument
    ) -> AbsolutePathCommand:
        """Return an absolute non-shorthand command

        .. versionadded:: 1.1
        """
        return self.to_absolute(prev)

    # The precision of the numbers when converting to string
    number_template = "{:.6g}"

    # Maps single letter path command to corresponding class
    # (filled at the bottom of file, when all classes already defined)
    _letter_to_class: Dict[str, Type[Any]] = {}

    @staticmethod
    def letter_to_class(letter):
        """Returns class for given path command letter"""
        return PathCommand._letter_to_class[letter]

    @property
    @abc.abstractmethod
    def args(self) -> List[float]:
        """Returns path command arguments as tuple of floats"""

    def control_points(
        self,
        first: ComplexLike,
        prev: ComplexLike,
        prev_prev: ComplexLike,
    ) -> Generator[Vector2d, None, None]:
        """Returns list of path command control points"""
        first, prev, prev_prev = complex(first), complex(prev), complex(prev_prev)
        yield from [Vector2d(i) for i in self.ccontrol_points(first, prev, prev_prev)]

    @abc.abstractmethod
    def ccontrol_points(
        self, first: complex, prev: complex, prev_prev: complex
    ) -> Tuple[complex, ...]:
        """Returns list of path command control points"""

    @classmethod
    def _argt(cls, sep):
        return sep.join([cls.number_template] * cls.nargs)

    def __str__(self):
        return f"{self.letter} {self._argt(' ').format(*self.args)}".strip()

    def __repr__(self):
        # pylint: disable=consider-using-f-string
        return "{{}}({})".format(self._argt(", ")).format(self.name, *self.args)

    def __eq__(self, other):
        previous = 0j
        if type(self) == type(other):  # pylint: disable=unidiomatic-typecheck
            return self.args == other.args
        if isinstance(other, tuple):
            return self.args == other
        if not isinstance(other, PathCommand):
            raise ValueError("Can't compare types")
        try:
            if self.is_relative == other.is_relative:
                return self.to_curve(previous) == other.to_curve(previous)
        except ValueError:
            pass
        return False

    @abc.abstractmethod
    def cend_point(self, first: complex, prev: complex) -> complex:
        """Complex version of end_point"""

    def end_point(self, first: ComplexLike, prev: ComplexLike) -> Vector2d:
        """Returns last control point of path command"""
        return Vector2d(self.cend_point(complex(first or 0), complex(prev or 0)))

    @abc.abstractmethod
    def update_bounding_box(
        self, first: complex, last_two_points: List[complex], bbox: BoundingBox
    ):
        """Enlarges given bbox to contain path element.

        Args:
            first (complex): first point of path. Required to calculate Z segment
            last_two_points (List[complex]): list with last two control points in abs
                coords.
            bbox (BoundingBox): bounding box to update
        """

    def to_curve(self, prev: ComplexLike, prev_prev: ComplexLike = 0) -> Curve:
        # pylint: disable=unused-argument
        """Convert command to :py:class:`Curve`

        Curve().to_curve() returns a copy
        """
        return NotImplemented

    def to_curves(self, prev: ComplexLike, prev_prev: ComplexLike = 0) -> List[Curve]:
        """Convert command to list of :py:class:`Curve` commands"""
        return [self.to_curve(prev, prev_prev)]

    def to_line(self, prev: ComplexLike) -> Line:
        # pylint: disable=unused-argument
        """Converts this segment to a line (copies if already a line)"""
        return NotImplemented

    @abc.abstractmethod
    def ccurve_points(
        self, first: complex, prev: complex, prev_prev: complex
    ) -> Tuple[complex, ...]:
        # pylint: disable=unused-argument
        """Converts the path element into a single cubic bezier"""

    # Derivation functionality

    def __check_t(self, t: Optional[float], allow_none=True):
        if not allow_none and (t is None and self.letter not in "zZmMlLhHvV"):
            raise ValueError("t=None only supported for Line-like commands")
        if t is not None and not 0 <= t <= 1:
            raise ValueError("t should be between 0 and 1")
        return t if t is not None else 0.0

    def cderivative(
        self,
        first: complex,
        prev: complex,
        prev_control: complex,
        t: Optional[float] = None,
        n: int = 1,
    ) -> complex:
        """Returns the nth derivative of the segment at t as a complex number.

        .. versionadded:: 1.4
        """
        if n < 1:
            raise ValueError("n should be a positive integer")
        # pylint: disable=protected-access
        return self._cderivative(first, prev, prev_control, self.__check_t(t), n)

    def derivative(
        self,
        first: ComplexLike,
        prev: ComplexLike,
        prev_control: ComplexLike,
        t: Optional[float] = None,
        n: int = 1,
    ) -> Vector2d:
        """Returns the nth derivative of the segment at t as a :class:`Vector2D`.

        .. versionadded:: 1.4
        """
        return Vector2d(
            self.cderivative(
                complex(first or 0),
                complex(prev or 0),
                complex(prev_control or 0),
                t,
                n,
            )
        )

    @abc.abstractmethod
    def _cderivative(
        self, first: complex, prev: complex, prev_control: complex, t: float, n: int = 1
    ) -> complex: ...

    def cunit_tangent(
        self,
        first: complex,
        prev: complex,
        prev_control: complex,
        t: Optional[float] = None,
    ) -> complex:
        """Returns the unit tangent of the segment at t as a complex number.

        ..versionadded:: 1.4
        """
        return self._cunit_tangent(first, prev, prev_control, self.__check_t(t))

    def unit_tangent(
        self,
        first: ComplexLike,
        prev: ComplexLike,
        prev_control: ComplexLike,
        t: Optional[float] = None,
    ) -> Vector2d:
        """Returns the unit tangent of the segment at t as a :class:`Vector2D`.

        ..versionadded:: 1.4
        """
        return Vector2d(
            self.cunit_tangent(
                complex(first or 0), complex(prev or 0), complex(prev_control or 0), t
            )
        )

    def _cunit_tangent(
        self, first: complex, prev: complex, prev_control: complex, t: float
    ) -> complex:
        dseg = self._cderivative(first, prev, t, 1)
        return dseg / abs(dseg)

    def cnormal(
        self,
        first: complex,
        prev: complex,
        prev_control: complex,
        t: Optional[float] = None,
    ) -> complex:
        """Returns the (right-hand-rule) normal vector of the segment at t as a complex
        number.

        ..versionadded:: 1.4
        """
        return self.cunit_tangent(first, prev, prev_control, t) * 1j

    def normal(
        self,
        first: ComplexLike,
        prev: ComplexLike,
        prev_control: ComplexLike,
        t: Optional[float] = None,
    ) -> Vector2d:
        """Returns the (right-hand-rule) normal vector of the segment at t as
        :class:`Vector2D`.

        ..versionadded:: 1.4
        """
        return Vector2d(
            self.cnormal(
                complex(first or 0), complex(prev or 0), complex(prev_control or 0), t
            )
        )

    def curvature(
        self,
        first: ComplexLike,
        prev: ComplexLike,
        prev_control: ComplexLike,
        t: Optional[float] = None,
    ) -> float:
        """Returns the curvature of the segment at t.

        ..versionadded:: 1.4
        """
        # pylint: disable=protected-access
        return self._curvature(
            complex(first or 0),
            complex(prev or 0),
            complex(prev_control or 0),
            self.__check_t(t),
        )

    @abc.abstractmethod
    def _curvature(
        self, first: complex, prev: complex, prev_control: complex, t: float
    ) -> float: ...

    # Point evaluation, splitting
    def cpoint(
        self, first: complex, prev: complex, prev_control: complex, t: float
    ) -> complex:
        """Returns the coordinates of the Bezier curve evaluated at t as complex number.

        .. versionadded:: 1.
        """
        return self._cpoint(first, prev, prev_control, self.__check_t(t, False))

    def point(
        self, first: ComplexLike, prev: ComplexLike, prev_control: ComplexLike, t: float
    ) -> Vector2d:
        """Returns the coordinates of the Bezier curve evaluated at t as :class:`Vector2d`.

        .. versionadded:: 1.4
        """
        return Vector2d(
            self.cpoint(
                complex(first or 0), complex(prev or 0), complex(prev_control or 0), t
            )
        )

    @abc.abstractmethod
    def _cpoint(
        self, first: complex, prev: complex, prev_control: complex, t: float
    ) -> complex: ...

    def split(
        self, first: ComplexLike, prev: ComplexLike, prev_control: ComplexLike, t: float
    ) -> Tuple[PathCommand, PathCommand]:
        """Returns two segments, whose union is this segment and which join at
        self.point(t).

        .. versionadded:: 1.4
        """
        # no simplification here, we want to preserve the original type
        return self._split(
            complex(first),
            complex(prev),
            complex(prev_control),
            self.__check_t(t, False),
        )

    @abc.abstractmethod
    def _split(
        self, first: complex, prev: complex, prev_control: complex, t: float
    ) -> Tuple[PathCommand, PathCommand]: ...

    # Line integration

    def length(
        self,
        first: ComplexLike,
        prev: ComplexLike,
        prev_control: ComplexLike,
        t0: float = 0,
        t1: float = 1,
        settings=LengthSettings(),
    ) -> float:
        """Returns the length of the segment between t0 and t1.

        .. versionadded:: 1.4
        """
        # pylint: disable=protected-access
        return self._length(
            complex(first),
            complex(prev),
            complex(prev_control),
            self.__check_t(t0, False),
            self.__check_t(t1, False),
            settings,
        )

    @abc.abstractmethod
    def _length(
        self,
        first: complex,
        prev: complex,
        prev_control: complex,
        t0: float = 0,
        t1: float = 1,
        settings=LengthSettings(),
    ) -> float: ...

    def ilength(
        self,
        first: ComplexLike,
        prev: ComplexLike,
        prev_control: ComplexLike,
        length: float,
        settings: ILengthSettings = ILengthSettings(),
    ):
        """Returns a float ``t``, such that ``self.length(0, t)`` is approximately
        ``length``.

        .. versionadded:: 1.4
        """
        # pylint: disable=protected-access
        return self._ilength(
            complex(first), complex(prev), complex(prev_control), length, settings
        )

    @abc.abstractmethod
    def _ilength(
        self,
        first: complex,
        prev: complex,
        prev_control: complex,
        length: float,
        settings: ILengthSettings = ILengthSettings(),
    ): ...


class RelativePathCommand(PathCommand):
    """
    Abstract base class for relative path commands.

    Implements most of methods of :py:class:`PathCommand` through
    conversion to :py:class:`AbsolutePathCommand`
    """

    @property
    def is_relative(self):
        return True

    @property
    def is_absolute(self):
        return False

    def to_relative(self, prev: ComplexLike) -> RelativePathCommand:
        return self.__class__(*self.args)

    def update_bounding_box(self, first, last_two_points, bbox):
        self.to_absolute(last_two_points[-1]).update_bounding_box(
            first, last_two_points, bbox
        )


class AbsolutePathCommand(PathCommand):
    """Absolute path command. Unlike :py:class:`RelativePathCommand` can be transformed
    directly."""

    @property
    def is_relative(self):
        return False

    @property
    def is_absolute(self):
        return True

    def to_absolute(self, prev: ComplexLike) -> AbsolutePathCommand:
        return self.__class__(*self.args)

    @abc.abstractmethod
    def transform(self, transform: Transform) -> AbsolutePathCommand:
        """Returns new transformed segment

        :param transform: a transformation to apply
        """

    def rotate(self, degrees: float, center: Vector2d) -> AbsolutePathCommand:
        """
        Returns new transformed segment

        :param degrees: rotation angle in degrees
        :param center: invariant point of rotation
        """
        return self.transform(Transform(rotate=(degrees, center[0], center[1])))

    def translate(self, dr: Vector2d) -> AbsolutePathCommand:
        """Translate or scale this path command by dr"""
        return self.transform(Transform(translate=dr))

    def scale(self, factor: Union[float, Tuple[float, float]]) -> AbsolutePathCommand:
        """Returns new transformed segment

        :param factor: scale or (scale_x, scale_y)
        """
        return self.transform(Transform(scale=factor))


class BezierArcComputationMixin:
    """Functionality that works the same way for Arcs, Cubic and Quadratic

    .. versionadded:: 1.4
    """

    _cpoint: Callable[[complex, complex, complex, float], complex]
    _cderivative: Callable[..., complex]
    poly: Callable[[complex, complex, complex, Optional[bool]], np.poly1d]

    def inv_arclength(self, prev, prev_control, s, settings=ILengthSettings()):
        """Ported from https://github.com/mathandy/svgpathtools/blob/19df25b99b405ec4fc7616b58384eca7879b6fd4/svgpathtools/path.py#L541
        (MIT Licensed)"""
        t_upper = 1
        t_lower = 0
        iteration = 0
        while iteration < settings.maxits:
            iteration += 1
            t = (t_lower + t_upper) / 2
            s_t = self._length(0j, prev, prev_control, t1=t, settings=settings)
            if abs(s_t - s) < settings.length_tol:
                return t
            elif s_t < s:  # t too small
                t_lower = t
            else:  # s < s_t, t too big
                t_upper = t
            if t_upper == t_lower:
                # warn("t is as close as a float can be to the correct value, "
                #        "but |s(t) - s| = {} > s_tol".format(abs(s_t-s)))
                return t
        raise Exception(f"Maximum iterations reached with s(t) - s = {s_t - s}.")

    def segment_length(
        self,
        start,
        end,
        start_point,
        end_point,
        pos_eval,
        settings=LengthSettings(),
        depth=0,
    ):
        """
        Ported from https://github.com/mathandy/svgpathtools/blob/19df25b99b405ec4fc7616b58384eca7879b6fd4/svgpathtools/path.py#L479
        (MIT licensed)
        # TODO better treatment of degenerate beziers from
        # https://github.com/linebender/kurbo/blob/c229a914d303c5989c9e6b1d766def2df27a8185/src/cubicbez.rs#L431
        """
        mid = (start + end) / 2
        mid_point = pos_eval(mid)
        length = abs(end_point - start_point)
        first_half = abs(mid_point - start_point)
        second_half = abs(end_point - mid_point)

        length2 = first_half + second_half
        if (length2 - length > settings.error) or (depth < settings.min_depth):
            # Calculate the length of each segment:
            depth += 1
            return self.segment_length(
                start, mid, start_point, mid_point, pos_eval, settings, depth
            ) + self.segment_length(
                mid, end, mid_point, end_point, pos_eval, settings, depth
            )
        # This is accurate enough.
        return length2

    def segment_curvature(self, prev: complex, prev_prev: complex, t: float):
        """Returns the curvature of the segment at t.

        Ported from https://github.com/mathandy/svgpathtools/blob/19df25b99b405ec4fc7616b58384eca7879b6fd4/svgpathtools/path.py#L386
        (MIT licensed)
        """

        dz = self._cderivative(0j, prev, prev_prev, t)
        ddz = self._cderivative(0j, prev, prev_prev, t, n=2)
        dx, dy = dz.real, dz.imag
        ddx, ddy = ddz.real, ddz.imag
        try:
            kappa = abs(dx * ddy - dy * ddx) / math.sqrt(dx * dx + dy * dy) ** 3
        except (ZeroDivisionError, FloatingPointError):
            # tangent vector is zero at t, use polytools to find limit
            p = self.poly(0j, prev, prev_prev, False)
            dp = p.deriv()
            ddp = dp.deriv()
            dx2, dy2 = np.real(dp), np.imag(dp)
            ddx2, ddy2 = np.real(ddp), np.imag(ddp)
            f2 = (dx2 * ddy2 - dy2 * ddx2) ** 2
            g2 = (dx2 * dx2 + dy2 * dy2) ** 3
            lim2 = rational_limit(f2, g2, t)
            if lim2 < 0:  # impossible, must be numerical error
                return 0
            kappa = math.sqrt(lim2)
        return kappa

    def _length(
        self,
        first: complex,
        prev: complex,
        prev_control: complex,
        t0=0,
        t1=1,
        settings=LengthSettings(),
    ) -> float:
        return self.segment_length(
            t0,
            t1,
            self._cpoint(first, prev, prev_control, t0),
            self._cpoint(first, prev, prev_control, t1),
            lambda t: self._cpoint(first, prev, prev_control, t),
            settings,
            0,
        )

    def _ilength(
        self,
        first: complex,
        prev: complex,
        prev_control: complex,
        length,
        settings=ILengthSettings(),
    ):
        return self.inv_arclength(prev, prev_control, length, settings)

    def _curvature(
        self, first: complex, prev: complex, prev_control: complex, t: float
    ) -> float:
        return self.segment_curvature(prev, prev_control, t)


class BezierComputationMixin:
    """Functionality that works the same for all Beziers (Quadratic, Cubic)

    .. versionadded:: 1.4
    """

    def _bpoints(self, prev, prev_prev):
        return (prev,) + self.ccontrol_points(0j, prev, prev_prev)

    def bezier_unit_tangent(self, prev, prev_prev, t):
        """Returns the unit tangent of the segment at t.

        Ported from https://github.com/mathandy/svgpathtools/blob/19df25b99b405ec4fc7616b58384eca7879b6fd4/svgpathtools/path.py#L348
        (MIT licensed)
        """

        dseg = self._cderivative(0j, prev, prev_prev, t)

        try:
            unit_tangent = dseg / abs(dseg)
        except (ZeroDivisionError, FloatingPointError):
            # This may be a removable singularity, if so we just need to compute
            # the limit.
            # Note: limit{{dseg / abs(dseg)} = sqrt(limit{dseg**2 / abs(dseg)**2})
            dseg_poly = self.poly(prev, prev_prev).deriv()
            dseg_abs_squared_poly = np.real(dseg_poly) ** 2 + np.imag(dseg_poly) ** 2
            try:
                unit_tangent = np.sqrt(
                    rational_limit(dseg_poly**2, dseg_abs_squared_poly, t)
                )
            except ValueError:
                bef = self.poly(prev, prev_prev).deriv()(t - 1e-4)
                aft = self.poly(prev, prev_prev).deriv()(t + 1e-4)
                mes = (
                    "Unit tangent appears to not be well-defined at "
                    f"t = {t}, \n"
                    f"seg.poly().deriv()(t - 1e-4) = {bef}\n"
                    f"seg.poly().deriv()(t + 1e-4) = {aft}"
                )
                raise ValueError(mes)
        return unit_tangent



#!/usr/bin/env python3
"""
Run the scour module on the svg output.
"""

import inkex
from inkex.localization import inkex_gettext as _

try:
    from packaging.version import Version
except ImportError:
    raise inkex.DependencyError(
        _(
            """Failed to import module 'packaging'.
Please make sure it is installed (e.g. using 'pip install packaging'
or 'sudo apt-get install python3-packaging') and try again.
"""
        )
    )

try:
    import scour
    from scour.scour import scourString
except ImportError:
    raise inkex.DependencyError(
        _(
            """Failed to import module 'scour'.
Please make sure it is installed (e.g. using 'pip install scour'
  or 'sudo apt-get install python3-scour') and try again.
"""
        )
    )


class ScourInkscape(inkex.OutputExtension):
    """Scour Inkscape Extension"""

    # Scour options
    def add_arguments(self, pars):
        pars.add_argument("--tab")
        pars.add_argument("--simplify-colors", type=inkex.Boolean, dest="simple_colors")
        pars.add_argument("--style-to-xml", type=inkex.Boolean)
        pars.add_argument(
            "--group-collapsing", type=inkex.Boolean, dest="group_collapse"
        )
        pars.add_argument("--create-groups", type=inkex.Boolean, dest="group_create")
        pars.add_argument("--enable-id-stripping", type=inkex.Boolean, dest="strip_ids")
        pars.add_argument("--shorten-ids", type=inkex.Boolean)
        pars.add_argument("--shorten-ids-prefix")
        pars.add_argument("--embed-rasters", type=inkex.Boolean)
        pars.add_argument(
            "--keep-unreferenced-defs", type=inkex.Boolean, dest="keep_defs"
        )
        pars.add_argument("--keep-editor-data", type=inkex.Boolean)
        pars.add_argument("--remove-metadata", type=inkex.Boolean)
        pars.add_argument("--strip-xml-prolog", type=inkex.Boolean)
        pars.add_argument("--set-precision", type=int, dest="digits")
        pars.add_argument("--indent", dest="indent_type")
        pars.add_argument("--nindent", type=int, dest="indent_depth")
        pars.add_argument("--line-breaks", type=inkex.Boolean, dest="newlines")
        pars.add_argument(
            "--strip-xml-space", type=inkex.Boolean, dest="strip_xml_space_attribute"
        )
        pars.add_argument("--protect-ids-noninkscape", type=inkex.Boolean)
        pars.add_argument("--protect-ids-list")
        pars.add_argument("--protect-ids-prefix")
        pars.add_argument("--enable-viewboxing", type=inkex.Boolean)
        pars.add_argument(
            "--enable-comment-stripping", type=inkex.Boolean, dest="strip_comments"
        )
        pars.add_argument("--renderer-workaround", type=inkex.Boolean)

        # options for internal use of the extension
        pars.add_argument("--scour-version")
        pars.add_argument("--scour-version-warn-old", type=inkex.Boolean)

    def save(self, stream):
        # version check if enabled in options
        if self.options.scour_version_warn_old:
            scour_version = scour.__version__
            scour_version_min = self.options.scour_version
            if Version(scour_version) < Version(scour_version_min):
                raise inkex.AbortExtension(
                    f"""
The extension 'Optimized SVG Output' is designed for Scour {scour_version_min} or later but you're
 using the older version Scour {scour_version}.

Note: You can permanently disable this message on the 'About' tab of the extension window."""
                )
        del self.options.scour_version
        del self.options.scour_version_warn_old

        # do the scouring
        stream.write(scourString(self.svg.tostring(), self.options).encode("utf8"))


if __name__ == "__main__":
    ScourInkscape().run()



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2010 Craig Marshall, craig9 [at] gmail.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
#
"""
This script slices an inkscape drawing along the guides, similarly to
the GIMP plugin called "guillotine". It can optionally export to the
same directory as the SVG file with the same name, but with a number
suffix. e.g.

/home/foo/drawing.svg

will export to:

/home/foo/drawing0.png
/home/foo/drawing1.png
/home/foo/drawing2.png
/home/foo/drawing3.png

etc.

"""

import os
import locale

import inkex
from inkex.command import inkscape
from inkex.localization import inkex_gettext as _


class Guillotine(inkex.EffectExtension):
    """Exports slices made using guides"""

    def add_arguments(self, pars):
        pars.add_argument("--directory", type=str, dest="directory")
        pars.add_argument("--image", type=str, dest="image")
        pars.add_argument("--ignore", type=inkex.Boolean, dest="ignore")

    def get_all_horizontal_guides(self):
        """
        Returns all horizontal guides as a list of floats stored as
        strings. Each value is the position from 0 in pixels.
        """
        for guide in self.svg.namedview.get_guides():
            if guide.is_horizontal:
                yield guide.raw_position.y

    def get_all_vertical_guides(self):
        """
        Returns all vertical guides as a list of floats stored as
        strings. Each value is the position from 0 in pixels.
        """
        for guide in self.svg.namedview.get_guides():
            if guide.is_vertical:
                yield guide.raw_position.x

    def get_horizontal_slice_positions(self):
        """
        Make a sorted list of all horizontal guide positions,
        including 0 and the document height, but not including
        those outside of the canvas
        """
        horizontals = [0.0]
        height = float(self.svg.viewbox_height)
        for y in self.get_all_horizontal_guides():
            if 0.0 < y <= height:
                horizontals.append(height - y)
        horizontals.append(height)
        return sorted(horizontals)

    def get_vertical_slice_positions(self):
        """
        Make a sorted list of all vertical guide positions,
        including 0 and the document width, but not including
        those outside of the canvas.
        """
        verticals = [0.0]
        width = float(self.svg.viewbox_width)
        for x in self.get_all_vertical_guides():
            if 0.0 < x <= width:
                verticals.append(x)
        verticals.append(width)
        return sorted(verticals)

    def get_slices(self):
        """
        Returns a list of all "slices" as denoted by the guides
        on the page. Each slice is really just a 4 element list of
        floats (stored as strings), consisting of the X and Y start
        position and the X and Y end position.
        """
        hs = self.get_horizontal_slice_positions()
        vs = self.get_vertical_slice_positions()
        # The --export-width argument is in viewport units
        hs = [self.svg.unit_to_viewport(i) for i in hs]
        vs = [self.svg.unit_to_viewport(j) for j in vs]
        slices = []
        for i in range(len(hs) - 1):
            for j in range(len(vs) - 1):
                slices.append([vs[j], hs[i], vs[j + 1], hs[i + 1]])
        return slices

    def get_filename_parts(self):
        """
        Attempts to get directory and image as passed in by the inkscape
        dialog. If the boolean ignore flag is set, then it will ignore
        these settings and try to use the settings from the export
        filename.
        """

        if not self.options.ignore:
            if self.options.image == "" or self.options.image is None:
                raise inkex.AbortExtension(_("Please enter an image name"))
            return self.options.directory, self.options.image
        else:
            """
            First get the export-filename from the document, if the
            document has been exported before (TODO: Will not work if it
            hasn't been exported yet), then uses this to return a tuple
            consisting of the directory to export to, and the filename
            without extension.
            """
            try:
                export_file = self.svg.get("inkscape:export-filename")
            except KeyError:
                raise inkex.AbortExtension(
                    _(
                        "To use the export hints option, you "
                        "need to have previously exported the document. "
                        "Otherwise no export hints exist!"
                    )
                )
            dirname, filename = os.path.split(export_file)
            filename = filename.rsplit(".", 1)[0]  # Without extension
            return dirname, filename

    def get_localised_string(self, name):
        return locale.format_string("%.f", float(name), 0)

    def export_slice(self, sli, filename):
        """
        Runs inkscape's command line interface and exports the image
        slice from the 4 coordinates in s, and saves as the filename
        given.
        """
        coords = ":".join([self.get_localised_string(dim) for dim in sli])
        inkex.errormsg(coords)
        inkscape(self.options.input_file, export_area=coords, export_filename=filename)

    def export_slices(self, slices):
        """
        Takes the slices list and passes each one with a calculated
        filename/directory into export_slice.
        """
        dirname, filename = self.get_filename_parts()
        # Remove some crusty extensions from name template
        if filename.endswith(".svg") or filename.endswith(".png"):
            filename = filename.rsplit(".", 1)[0]
        if "{" not in filename:
            filename += "_{}"

        dirname = os.path.abspath(
            os.path.expanduser(os.path.expandvars(dirname or "./"))
        )
        if not os.path.isdir(dirname):
            os.makedirs(dirname)

        output_files = []
        for i, slico in enumerate(slices):
            fname = os.path.join(dirname, filename.format(i) + ".png")
            output_files.append(fname)
            self.export_slice(slico, fname)

        self.debug(
            _("The sliced bitmaps have been saved as:\n\n") + "\n".join(output_files)
        )

    def effect(self):
        self.export_slices(self.get_slices())


if __name__ == "__main__":
    Guillotine().run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright 2018 Martin Owens <doctormo@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>
#
"""
Inkscape Extensions Manager, Graphical User Interface (Gtk3)
"""

import sys
import logging
import warnings

from argparse import ArgumentParser

warnings.filterwarnings("ignore")


def run(args):
    # Late imports to catch import errors.
    from inkman.targets import TARGETS
    from inkman.gui import ManagerApp
    from inkman.utils import get_inkscape_version

    arg_parser = ArgumentParser(description=__doc__)
    arg_parser.add_argument("input_file", nargs="?")
    arg_parser.add_argument(
        "--target",
        default=TARGETS[0].category,
        choices=[t.category for t in TARGETS],
    )
    arg_parser.add_argument("--for-version", default=None)
    options = arg_parser.parse_args(args)
    version = options.for_version or get_inkscape_version()
    try:
        ManagerApp(
            start_loop=True,
            version=version,
            target=[t for t in TARGETS if t.category == options.target][0],
        )
    except KeyboardInterrupt:
        logging.error("User Interputed")
    logging.debug("Exiting Application")


def recovery_run(args):
    try:
        run(args)
    except Exception:
        from inkman.backfoot import attempt_to_recover

        attempt_to_recover()


if __name__ == "__main__":
    recovery_run(sys.argv[1:])



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2005, 2007 Aaron Spike, aaron@ekips.org
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
This extension either adds nodes to a path so that

  No segment is longer than a maximum value OR that each segment is divided
  into a given number of equal segments.

"""

import math
from typing import cast
import inkex

from inkex import PathElement


class AddNodes(inkex.EffectExtension):
    """Extension to split a path by adding nodes to it"""

    def add_arguments(self, pars):
        pars.add_argument(
            "--segments",
            type=int,
            default=2,
            help="Number of segments to divide the path into",
        )
        pars.add_argument(
            "--max",
            type=float,
            default=10.0,
            help="Number of segments to divide the path into",
        )
        pars.add_argument(
            "--method", default="bymax", help="The kind of division to perform"
        )
        pars.add_argument(
            "--unit", default="px", help="Unit for maximum segment length"
        )

    def effect(self):
        maxlen = self.svg.viewport_to_unit(f"{self.options.max}{self.options.unit}")
        for node in self.svg.selection.filter(PathElement):
            new = inkex.Path()
            path = cast(inkex.Path, node.path)
            for sub in path.proxy_iterator():
                if sub.letter in "mM":
                    # Add unchanged.
                    new.append(sub.command)
                    continue
                length = sub.length()
                if length < 1e-12:
                    # Zero-length subpaths
                    new.append(sub.command)
                    continue
                # compute number of splits
                if self.options.method == "bynum":
                    splits = self.options.segments
                else:
                    splits = math.ceil(length / maxlen)
                to_split = sub
                # We first split the command at len * 1/splits.
                # The first part is appended, the second part is split
                # again at then len * 1/splits and so on.

                for _ in range(splits - 1):
                    result = to_split.split(to_split.ilength(length / splits))
                    new.append(result[0].command)
                    to_split = result[1]
                new.append(to_split.command)

            node.path = new


if __name__ == "__main__":
    AddNodes().run()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2020 Martin Owens <doctormo@gmail.com>
#                    Sergei Izmailov <sergei.a.izmailov@gmail.com>
#                    Thomas Holder <thomas.holder@schrodinger.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# pylint: disable=arguments-differ
"""
Provide extra utility to each svg element type specific to its type.

This is useful for having a common interface for each element which can
give path, transform, and property access easily.
"""

from __future__ import annotations

from copy import deepcopy
from typing import Any, Tuple, Optional, overload, TypeVar, List, Callable
from lxml import etree
import re

from ..interfaces.IElement import IBaseElement, ISVGDocumentElement

from ..base import SvgOutputMixin
from ..paths import Path
from ..styles import Style, Classes, StyleValue
from ..transforms import Transform, BoundingBox
from ..utils import FragmentError
from ..units import convert_unit, render_unit, parse_unit
from ._utils import ChildToProperty, NSS, addNS, removeNS, splitNS
from ..properties import (
    _ShorthandValueConverter,
    _get_tokens_from_value,
    all_properties,
)
from ._selected import ElementList
from ._parser import NodeBasedLookup, SVG_PARSER

T = TypeVar("T", bound="BaseElement")  # pylint: disable=invalid-name


class BaseElement(IBaseElement):
    """Provide automatic namespaces to all calls"""

    # pylint: disable=too-many-public-methods

    def __init_subclass__(cls):
        if cls.tag_name:
            NodeBasedLookup.register_class(cls)

    @classmethod
    def is_class_element(  # pylint: disable=unused-argument
        cls, elem: etree.Element
    ) -> bool:
        """Hook to do more restrictive check in addition to (ns,tag) match

        .. versionadded:: 1.2
            The function has been made public.
        """
        return True

    tag_name = ""

    @property
    def TAG(self):  # pylint: disable=invalid-name
        """Return the tag_name without NS"""
        if not self.tag_name:
            return removeNS(super().tag)[-1]
        return removeNS(self.tag_name)[-1]

    @classmethod
    def new(cls, *children, **attrs):
        """Create a new element, converting attrs values to strings."""
        obj = cls(*children)
        obj.update(**attrs)
        return obj

    NAMESPACE = property(lambda self: splitNS(self.tag_name)[0])
    """Get namespace of element"""

    PARSER = SVG_PARSER
    """A reference to the :attr:`inkex.elements._parser.SVG_PARSER`"""
    WRAPPED_ATTRS = (
        # (prop_name, [optional: attr_name], cls)
        ("transform", Transform),
        ("style", Style),
        ("classes", "class", Classes),
    )  # type: Tuple[Tuple[Any, ...], ...]
    """A list of attributes that are automatically converted to objects."""

    # We do this because python2 and python3 have different ways
    # of combining two dictionaries that are incompatible.
    # This allows us to update these with inheritance.
    @property
    def wrapped_attrs(self):
        """Map attributes to property name and wrapper class"""
        return {row[-2]: (row[0], row[-1]) for row in self.WRAPPED_ATTRS}

    @property
    def wrapped_props(self):
        """Map properties to attribute name and wrapper class"""
        return {row[0]: (row[-2], row[-1]) for row in self.WRAPPED_ATTRS}

    typename = property(lambda self: type(self).__name__)
    """Type name of the element"""

    xml_path = property(lambda self: self.getroottree().getpath(self))
    """XPath representation of the element in its tree

    .. versionadded:: 1.1
    """

    desc = ChildToProperty("svg:desc", prepend=True)
    """The element's long-form description (for accessibility purposes)

    .. versionadded:: 1.1
    """

    title = ChildToProperty("svg:title", prepend=True)
    """The element's short-form description (for accessibility purposes)

    .. versionadded:: 1.1
    """

    _root: Optional[ISVGDocumentElement] = None

    def __getattr__(self, name):
        """Get the attribute, but load it if it is not available yet"""
        if name in self.wrapped_props:
            (attr, cls) = self.wrapped_props[name]

            # The reason we do this here and not in _init is because lxml
            # is inconsistent about when elements are initialized.
            # So we make this a lazy property.
            def _set_attr(new_item):
                if new_item:
                    self.set(attr, str(new_item))
                else:
                    self.attrib.pop(attr, None)  # pylint: disable=no-member

            # pylint: disable=no-member
            value = cls(self.attrib.get(attr, None), callback=_set_attr, element=self)
            setattr(self, name, value)
            return value
        raise AttributeError(f"Can't find attribute {self.typename}.{name}")

    def __setattr__(self, name, value):
        """Set the attribute, update it if needed"""
        if name in self.wrapped_props:
            (attr, cls) = self.wrapped_props[name]
            # Don't call self.set or self.get (infinite loop)
            if value:
                if not isinstance(value, cls):
                    value = cls(value)
                self.attrib[attr] = str(value)
            else:
                self.attrib.pop(attr, None)  # pylint: disable=no-member
        else:
            super().__setattr__(name, value)

    def get(self, attr, default=None):
        """Get element attribute named, with addNS support."""
        if attr in self.wrapped_attrs:
            (prop, _) = self.wrapped_attrs[attr]
            value = getattr(self, prop, None)
            # We check the boolean nature of the value, because empty
            # transformations and style attributes are equiv to not-existing
            ret = str(value) if value else default
            return ret
        return super().get(addNS(attr), default)

    def set(self, attr, value):
        """Set element attribute named, with addNS support"""
        if attr == "id" and value is not None:
            try:
                oldid = self.get("id", None)
                if oldid is not None and oldid in self.root.ids:
                    self.root.ids.pop(oldid)
                if value in self.root.ids:
                    raise ValueError(f"ID {value} already exists in this document")
                self.root.ids[value] = self
            except FragmentError:  # element is unrooted
                pass
        if attr in self.wrapped_attrs:
            # Always keep the local wrapped class up to date.
            (prop, cls) = self.wrapped_attrs[attr]
            setattr(self, prop, cls(value))
            value = getattr(self, prop)
            if not value:
                return
        if value is None:
            self.attrib.pop(addNS(attr), None)  # pylint: disable=no-member
        else:
            value = str(value)
            super().set(addNS(attr), value)

    def update(self, **kwargs):
        """
        Update element attributes using keyword arguments

        Note: double underscore is used as namespace separator,
        i.e. "namespace__attr" argument name will be treated as "namespace:attr"

        :param kwargs: dict with name=value pairs
        :return: self
        """
        for name, value in kwargs.items():
            self.set(name, value)
        return self

    def pop(self, attr, default=None):
        """Delete/remove the element attribute named, with addNS support."""
        if attr in self.wrapped_attrs:
            # Always keep the local wrapped class up to date.
            (prop, cls) = self.wrapped_attrs[attr]
            value = getattr(self, prop)
            setattr(self, prop, cls(None))
            return value
        return self.attrib.pop(addNS(attr), default)  # pylint: disable=no-member

    @overload
    def add(
        self, child1: BaseElement, child2: BaseElement, *children: BaseElement
    ) -> Tuple[BaseElement]: ...

    @overload
    def add(self, child: T) -> T: ...

    def add(self, *children):
        """
        Like append, but will do multiple children and will return
        children or only child
        """
        for child in children:
            self.append(child)
        return children if len(children) != 1 else children[0]

    def tostring(self):
        """Return this element as it would appear in an svg document"""
        # This kind of hack is pure madness, but etree provides very little
        # in the way of fragment printing, preferring to always output valid xml

        svg = SvgOutputMixin.get_template(width=0, height=0).getroot()
        svg.append(self.copy())
        return svg.tostring().split(b">\n    ", 1)[-1][:-6]

    def set_random_id(
        self,
        prefix: Optional[str] = None,
        size: Optional[int] = None,
        backlinks: bool = False,
        blacklist: Optional[List[str]] = None,
    ):
        """Sets the id attribute if it is not already set.

        The id consists of a prefix and an appended random integer of length size.

        Args:
            prefix (str, optional): the prefix of the new ID. Defaults to the tag name.
            size (Optional[int], optional): number of digits of the second part of the
                id. If None, the length is chosen based on the amount of existing
                objects. Defaults to None.

                .. versionchanged:: 1.2
                    The default of this value has been changed from 4 to None.
            backlinks (bool, optional): Whether to update the links in existing objects
                that reference this element. Defaults to False.
            blacklist (List[str], optional): An additional list of ids that are not
                allowed to be used. This is useful when bulk inserting objects.
                Defaults to None.

                .. versionadded:: 1.2
        """
        prefix = str(self) if prefix is None else prefix
        self.set_id(
            self.root.get_unique_id(prefix, size=size, blacklist=blacklist),
            backlinks=backlinks,
        )

    def set_random_ids(
        self,
        prefix: Optional[str] = None,
        levels: int = -1,
        backlinks: bool = False,
        blacklist: Optional[List[str]] = None,
    ):
        """Same as set_random_id, but will apply also to children

        The id consists of a prefix and an appended random integer of length size.

        Args:
            prefix (str, optional): the prefix of the new ID. Defaults to the tag name.
            levels (int, optional): the depth of the tree traversion, if negative, no
                limit is imposed. Defaults to -1.
            backlinks (bool, optional): Whether to update the links in existing objects
                that reference this element. Defaults to False.
            blacklist (List[str], optional): An additional list of ids that are not
                allowed to be used. This is useful when bulk inserting objects.
                Defaults to None.

                .. versionadded:: 1.2
        """
        self.set_random_id(prefix=prefix, backlinks=backlinks, blacklist=blacklist)
        if levels != 0:
            for child in self:
                if hasattr(child, "set_random_ids"):
                    child.set_random_ids(
                        prefix=prefix, levels=levels - 1, backlinks=backlinks
                    )

    eid = property(lambda self: self.get_id())
    """Property to access the element's id; will set a new unique id if not set."""

    def get_id(self, as_url=0) -> str:
        """Get the id for the element, will set a new unique id if not set.

        as_url - If set to 1, returns #{id} as a string
                 If set to 2, returns url(#{id}) as a string

        Args:
            as_url (int, optional):
                - If set to 1, returns #{id} as a string
                - If set to 2, returns url(#{id}) as a string.

                Defaults to 0.

                .. versionadded:: 1.1

        Returns:
            str: formatted id
        """
        if "id" not in self.attrib:
            self.set_random_id(self.TAG)
        eid = self.get("id")
        if as_url > 0:
            eid = "#" + eid
        if as_url > 1:
            eid = f"url({eid})"
        return eid

    def set_id(self, new_id, backlinks=False):
        """Set the id and update backlinks to xlink and style urls if needed"""
        old_id = self.get("id", None)
        self.set("id", new_id)
        if backlinks and old_id:
            for elem in self.root.getElementsByHref(old_id):
                elem.href = self
            for attr in ["clip-path", "mask"]:
                for elem in self.root.getElementsByHref(old_id, attribute=attr):
                    elem.set(attr, self.get_id(2))
            for elem in self.root.getElementsByStyleUrl(old_id):
                elem.style.update_urls(old_id, new_id)

    @property
    def root(self) -> ISVGDocumentElement:
        """Get the root document element from any element descendent"""
        if self._root is not None:
            return self._root

        root, parent = self, self
        while parent is not None:
            root, parent = parent, parent.getparent()

        if not isinstance(root, ISVGDocumentElement):
            raise FragmentError("Element fragment does not have a document root!")

        self._root = root
        return root

    def get_or_create(self, xpath, nodeclass=None, prepend=False):
        """Get or create the given xpath, pre/append new node if not found.

        .. versionchanged:: 1.1
            The ``nodeclass`` attribute is optional; if not given, it is looked up
            using :func:`~inkex.elements._parser.NodeBasedLookup.find_class`
        """
        node = self.findone(xpath)
        if node is None:
            if nodeclass is None:
                nodeclass = NodeBasedLookup.find_class(xpath)
            node = nodeclass()
            if prepend:
                self.insert(0, node)
            else:
                self.append(node)
        return node

    def descendants(self):
        """Walks the element tree and yields all elements, parent first

        .. versionchanged:: 1.1
            The ``*types`` attribute was removed

        """

        return ElementList(
            self.root,
            [
                element
                for element in self.iter()
                if isinstance(element, (BaseElement, str))
            ],
        )

    def ancestors(self, elem=None, stop_at=()):
        """
        Walk the parents and yield all the ancestor elements, parent first

        Args:
            elem (BaseElement, optional): If provided, it will stop at the last common
                ancestor. Defaults to None.

                .. versionadded:: 1.1

            stop_at (tuple, optional): If provided, it will stop at the first parent
                that is in this list. Defaults to ().

                .. versionadded:: 1.1

        Returns:
            ElementList: list of ancestors
        """

        try:
            return ElementList(self.root, self._ancestors(elem=elem, stop_at=stop_at))
        except FragmentError:
            return ElementList(self, self._ancestors(elem=elem, stop_at=stop_at))

    def _ancestors(self, elem, stop_at):
        if isinstance(elem, BaseElement):
            stop_at = list(elem.ancestors())
        for parent in self.iterancestors():
            yield parent
            if parent in stop_at:
                break

    def backlinks(self, *types):
        """Get elements which link back to this element, like ancestors but via
        xlinks"""
        if not types or isinstance(self, types):
            yield self
        my_id = self.get("id")
        if my_id is not None:
            elems = list(self.root.getElementsByHref(my_id)) + list(
                self.root.getElementsByStyleUrl(my_id)
            )
            for elem in elems:
                if hasattr(elem, "backlinks"):
                    for child in elem.backlinks(*types):
                        yield child

    def xpath(self, pattern, namespaces=NSS):  # pylint: disable=dangerous-default-value
        """Wrap xpath call and add svg namespaces"""
        return super().xpath(pattern, namespaces=namespaces)

    def findall(self, pattern, namespaces=NSS):  # pylint: disable=dangerous-default-value
        """Wrap findall call and add svg namespaces"""
        return super().findall(pattern, namespaces=namespaces)

    def findone(self, xpath):
        """Gets a single element from the given xpath or returns None"""
        el_list = self.xpath(xpath)
        return el_list[0] if el_list else None

    def delete(self):
        """Delete this node from it's parent node"""
        if self.getparent() is not None:
            self.getparent().remove(self)

    def remove_all(self, *types):
        """Remove all children or child types

        .. versionadded:: 1.1
        """
        types = tuple(NodeBasedLookup.find_class(t) for t in types)
        for child in self:
            if not types or isinstance(child, types):
                self.remove(child)

    def replace_with(self, elem):
        """Replace this element with the given element"""
        self.getparent().replace(self, elem)

        if not elem.get("id") and self.get("id"):
            elem.set("id", self.get("id"))
        if not elem.label and self.label:
            elem.label = self.label
        return elem

    def copy(self):
        """Make a copy of the element and return it"""
        elem = deepcopy(self)
        elem.set("id", None)
        return elem

    def duplicate(self):
        """Like copy(), but the copy stays in the tree and sets a random id on the
        duplicate.

        .. versionchanged:: 1.2
            A random id is also set on all the duplicate's descendants
        """
        elem = self.copy()
        self.addnext(elem)
        elem.set_random_ids()
        return elem

    def __str__(self):
        # We would do more here, but lxml is VERY unpleasant when it comes to
        # namespaces, basically over printing details and providing no
        # suppression mechanisms to turn off xml's over engineering.
        return str(self.tag).split("}", maxsplit=1)[-1]

    @property
    def href(self):
        """Returns the referred-to element if available

        .. versionchanged:: 1.1
            A setter for href was added.
        """
        ref = self.get("href") or self.get("xlink:href")
        if not ref:
            return None
        return self.root.getElementById(ref.strip("#"))

    @href.setter
    def href(self, elem):
        """Set the href object"""
        if isinstance(elem, BaseElement):
            elem = elem.get_id()
        if self.get("href"):
            self.set("href", "#" + elem)
        else:
            self.set("xlink:href", "#" + elem)

    @property
    def label(self):
        """Returns the inkscape label"""
        return self.get("inkscape:label", None)

    @label.setter
    def label(self, value):
        """Sets the inkscape label"""
        self.set("inkscape:label", str(value))

    def is_sensitive(self):
        """Return true if this element is sensitive in Inkscape

        .. versionadded:: 1.1
        """
        return self.get("sodipodi:insensitive", None) != "true"

    def set_sensitive(self, sensitive=True):
        """Set the sensitivity of the element/layer

        .. versionadded:: 1.1
        """
        # Sensitive requires None instead of 'false'
        self.set("sodipodi:insensitive", ["true", None][sensitive])

    @property
    def unit(self):
        """Return the unit being used by the owning document, cached

        .. versionadded:: 1.1
        """
        try:
            return self.root.unit
        except FragmentError:
            return "px"  # Don't cache.

    @staticmethod
    def to_dimensional(value, to_unit="px"):
        """Convert a value given in user units (px) the given unit type

        .. versionadded:: 1.2
        """
        return convert_unit(value, to_unit)

    @staticmethod
    def to_dimensionless(value):
        """Convert a length value into user units (px)

        .. versionadded:: 1.2
        """
        return convert_unit(value, "px")

    def uutounit(self, value, to_unit="px"):
        """Convert a unit value to a given unit. If the value does not have a unit,
        "Document" units are assumed. "Document units" are an Inkscape-specific concept.
        For most use-cases, :func:`to_dimensional` is more appropriate.

        .. versionadded:: 1.1
        """
        return convert_unit(value, to_unit, default=self.unit)

    def unittouu(self, value):
        """Convert a unit value into document units. "Document unit" is an
        Inkscape-specific concept. For most use-cases, :func:`viewport_to_unit` (when
        the size of an object given in viewport units is needed) or
        :func:`to_dimensionless` (when the equivalent value without unit is needed) is
        more appropriate.

        .. versionadded:: 1.1
        """
        return convert_unit(value, self.unit)

    def unit_to_viewport(self, value, unit="px"):
        """Converts a length value to viewport units, as defined by the width/height
        element on the root (i.e. applies the equivalent transform of the viewport)

        .. versionadded:: 1.2
        """
        return self.to_dimensional(
            self.to_dimensionless(value) * self.root.equivalent_transform_scale, unit
        )

    def viewport_to_unit(self, value, unit="px"):
        """Converts a length given on the viewport to the specified unit in the user
        coordinate system

        .. versionadded:: 1.2
        """
        return self.to_dimensional(
            self.to_dimensionless(value) / self.root.equivalent_transform_scale, unit
        )

    def add_unit(self, value):
        """Add document unit when no unit is specified in the string.

        .. versionadded:: 1.1
        """
        return render_unit(value, self.unit)

    def cascaded_style(self):
        """Returns the cascaded style of an element (all rules that apply the element
        itself), based on the stylesheets, the presentation attributes and the inline
        style using the respective specificity of the style.

        see https://www.w3.org/TR/CSS22/cascade.html#cascading-order

        .. versionadded:: 1.2

        Returns:
            Style: the cascaded style

        """
        return Style.cascaded_style(self)

    def specified_style(self):
        """Returns the specified style of an element, i.e. the cascaded style +
        inheritance, see https://www.w3.org/TR/CSS22/cascade.html#specified-value.

        Returns:
            Style: the specified style

        .. versionadded:: 1.2
        """
        return Style.specified_style(self)

    def get_computed_style(self, key):
        """Returns the computed style value with respect to an element, i.e. the
        cascaded style + inheritance, see
        https://www.w3.org/TR/CSS22/cascade.html#computed-value.

        This is more efficient if only few style values per element are queried. If
        many attributes are queried, use :func:`specified_style`.

        .. versionadded:: 1.4
        """
        return Style._get_style(key, self)

    def presentation_style(self):
        """Return presentation attributes of an element as style

        .. versionadded:: 1.2
        """
        style = Style()
        for key in self.keys():
            if (
                key in all_properties
                # Shorthands cannot be set by presentation attributes
                and not isinstance(
                    all_properties[key].converter, _ShorthandValueConverter
                )
                and all_properties[key].presentation
            ):
                style[key] = StyleValue(_get_tokens_from_value(self.attrib[key]))
        return style

    def composed_transform(self, other=None):
        """Calculate every transform down to the other element
        if none specified the transform is to the root document element
        """
        parent = self.getparent()
        if parent is not other and isinstance(parent, BaseElement):
            return parent.composed_transform(other) @ self.transform
        return self.transform

    def _add_to_tree_callback(self, element):
        try:
            element._root = self._root
            self.root.add_to_tree_callback(element)
        except (FragmentError, AttributeError):
            pass

    @staticmethod
    def _remove_from_tree_callback(oldtree, element):
        try:
            oldtree.root.remove_from_tree_callback(element)
        except (FragmentError, AttributeError):
            pass

    def __element_adder(
        self, element: BaseElement, add_func: Callable[[BaseElement], None]
    ):
        BaseElement._remove_from_tree_callback(element, element)
        # Make sure that we have an ID cache before adding the element,
        # otherwise we will try to add this element twice to the cache
        try:
            self.root.ids
        except FragmentError:
            pass
        try:
            add_func(element)
        except Exception as err:
            BaseElement._add_to_tree_callback(element, element)
            raise err
        self._add_to_tree_callback(element)

    # Overrides to keep track of styles and IDs
    def addnext(self, element):
        self.__element_adder(element, super(etree.ElementBase, self).addnext)

    def addprevious(self, element):
        self.__element_adder(element, super(etree.ElementBase, self).addprevious)

    def append(self, element):
        self.__element_adder(element, super(etree.ElementBase, self).append)

    def clear(self, keep_tail=False):
        subelements = iter(self)
        old_id = self.get("id", None)
        super().clear(keep_tail)
        for element in subelements:
            BaseElement._remove_from_tree_callback(self, element)
        if old_id is not None and old_id in self.root.ids:
            self.root.ids.pop(old_id)

    def extend(self, elements):
        for element in elements:
            BaseElement._remove_from_tree_callback(element, element)
        try:
            self.root.ids
        except FragmentError:
            pass
        try:
            super().extend(elements)
        except Exception as err:
            for element in elements:
                BaseElement._add_to_tree_callback(element, element)
            raise err
        for element in elements:
            self._add_to_tree_callback(element)

    def insert(self, index, element):
        self.__element_adder(
            element,
            lambda element: super(etree.ElementBase, self).insert(index, element),
        )

    def remove(self, element):
        super().remove(element)
        BaseElement._remove_from_tree_callback(self, element)

    def replace(self, old_element, new_element):
        def replacer(new_element):
            super(etree.ElementBase, self).replace(old_element, new_element)
            BaseElement._remove_from_tree_callback(self, old_element)

        self.__element_adder(new_element, replacer)


NodeBasedLookup.default = BaseElement


class ShapeElement(BaseElement):
    """Elements which have a visible representation on the canvas"""

    @property
    def path(self) -> Path:
        """Gets the outline or path of the element, this may be a simple bounding box"""
        return self.get_path()

    @path.setter
    def path(self, path):
        self.set_path(path)

    @property
    def clip(self):
        """Gets the clip path element (if any). May be set through CSS.

        .. versionadded:: 1.1
        """
        return self.get_computed_style("clip-path")

    @clip.setter
    def clip(self, elem):
        self.set("clip-path", elem.get_id(as_url=2))

    def get_path(self) -> Path:
        """Generate a path for this object which can inform the bounding box"""
        raise NotImplementedError(
            f"Path should be provided by svg elem {self.typename}."
        )

    def set_path(self, path):
        """Set the path for this object (if possible)"""
        raise AttributeError(
            f"Path can not be set on this element: {self.typename} <- {path}."
        )

    def to_path_element(self):
        """Replace this element with a path element"""
        from ._polygons import PathElement

        elem = PathElement()
        elem.path = self.path
        elem.style = self.effective_style()
        elem.transform = self.transform
        return elem

    def effective_style(self):
        """Without parent styles, what is the effective style is"""
        return self.style

    def bounding_box(self, transform=None):
        # type: (Optional[Transform]) -> Optional[BoundingBox]
        """BoundingBox of the shape

        .. versionchanged:: 1.1
            result adjusted for element's clip path if applicable.
        """
        shape_box = self.shape_box(transform)
        clip = self.clip
        if clip is None or shape_box is None:
            return shape_box
        return shape_box & clip.bounding_box(Transform(transform) @ self.transform)

    def shape_box(self, transform=None):
        # type: (Optional[Transform]) -> Optional[BoundingBox]
        """BoundingBox of the unclipped shape

        .. versionadded:: 1.1
            Previous :func:`bounding_box` function, returning the bounding box
            without computing the effect of a possible clip.
        """
        path = self.path.to_absolute()
        if transform is True:
            path = path.transform(self.composed_transform())
        else:
            path = path.transform(self.transform)
            if transform:  # apply extra transformation
                path = path.transform(transform)
        return path.bounding_box()

    def is_visible(self):
        """Returns false if this object is invisible

        .. versionchanged:: 1.3
            rely on cascaded_style() to include CSS and presentation attributes
            include `visibility` attribute with check for inherit
            include ancestors

        .. versionadded:: 1.1
        """
        return self._is_visible()

    def _is_visible(self, inherit_visibility=True):
        # iterate over self and ancestors
        # This does not use :func:`get_computed_style` but its own iteration
        # logic to avoid duplicate evaluation of styles: a child is also invisible
        # if the parent has opacity:0, but opacity is not inherited - so we need
        # to check the specified style of all parents and ignore inheritance
        # altogether
        for element in [self] + list(self.ancestors()):
            get_style = element.cascaded_style().get
            # case display:none
            if get_style("display", "inline") == "none":
                return False
            # case opacity:0
            if not float(get_style("opacity", 1.0)):
                return False
            # only check if child's visibility is inherited
            if inherit_visibility:
                # case visibility:hidden
                if get_style("visibility", "inherit") in (
                    "hidden",
                    "collapse",
                ):
                    return False
                # case visibility: not inherit
                elif get_style("visibility", "inherit") != "inherit":
                    inherit_visibility = False

        return True

    def get_line_height_uu(self):
        """Returns the specified value of line-height, in user units

        .. versionadded:: 1.1
        """
        style = self.specified_style()
        font_size = style("font-size")  # already in uu
        line_height = style("line-height")
        parsed = parse_unit(line_height)
        if parsed is None:
            return font_size * 1.2
        if parsed[1] == "%":
            return font_size * parsed[0] * 0.01
        return self.to_dimensionless(line_height)


class ViewboxMixin:
    """Mixin for elements with viewboxes, such as <svg>, <marker>"""

    def parse_viewbox(self, vbox: Optional[str]) -> Optional[List[float]]:
        """Parses a viewbox. If an error occurs during parsing,
        (0, 0, 0, 0) is returned. If the viewbox is None, None is returned.

        .. versionadded:: 1.3
        """
        if vbox is not None and isinstance(vbox, str):
            try:
                result = [float(unit) for unit in re.split(r",\s*|\s+", vbox)]
            except ValueError:
                result = []
            if len(result) != 4:
                result = [0, 0, 0, 0]
            return result
        return None



# coding=utf-8
#
# Copyright (C) 2018 Martin Owens <doctormo@gmail.com>
# Copyright (C) 2023 Jonathan Neuhauser <jonathan.neuhauser@outlook.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
Path and CubicSuperPath classes
"""

from __future__ import annotations

import re
import copy
import warnings
from cmath import isclose

from typing import Optional, Tuple, List, TypeVar, Iterator, Callable, Union
from ..transforms import (
    Transform,
    BoundingBox,
    Vector2d,
    ComplexLike,
)
from ..utils import strargs

from .lines import Line, Move, move, ZoneClose, zoneClose
from .curves import Curve
from .interfaces import (
    ILengthSettings,
    LengthSettings,
    PathCommand,
    AbsolutePathCommand,
)

Pathlike = TypeVar("Pathlike", bound="PathCommand")
AbsolutePathlike = TypeVar("AbsolutePathlike", bound="AbsolutePathCommand")

LEX_REX = re.compile(r"([MLHVCSQTAZmlhvcsqtaz])([^MLHVCSQTAZmlhvcsqtaz]*)")


class InvalidPath(ValueError):
    """Raised when given an invalid path string"""


class Path(list):
    """A list of segment commands which combine to draw a shape"""

    callback: Optional[Callable] = None

    class PathCommandProxy:
        """
        A handy class for Path traverse and coordinate access

        Reduces number of arguments in user code compared to bare
        :class:`PathCommand` methods
        """

        def __init__(
            self,
            command: PathCommand,
            first_point: ComplexLike,
            previous_end_point: ComplexLike,
            prev2_control_point: ComplexLike,
        ):
            self.command = command
            self.cfirst_point = complex(first_point)
            self.cprevious_end_point = complex(previous_end_point)
            self.cprev2_control_point = complex(prev2_control_point)

        @property
        def first_point(self) -> Vector2d:
            """First point of the current subpath"""
            return Vector2d(self.cfirst_point)

        @property
        def previous_end_point(self) -> Vector2d:
            """End point of the previous command"""
            return Vector2d(self.cprevious_end_point)

        @property
        def prev2_control_point(self) -> Vector2d:
            """Last control point of the previous command"""
            return Vector2d(self.cprev2_control_point)

        @property
        def name(self) -> str:
            """The full name of the segment (i.e. Line, Arc, etc)"""
            return self.command.name

        @property
        def letter(self) -> str:
            """The single letter representation of this command (i.e. L, A, etc)"""
            return self.command.letter

        @property
        def next_command(self):
            """The implicit next command."""
            return self.command.next_command

        @property
        def is_relative(self) -> bool:
            """Whether the command is defined in relative coordinates, i.e. relative to
            the previous endpoint (lower case path command letter)"""
            return self.command.is_relative

        @property
        def is_absolute(self) -> bool:
            """Whether the command is defined in absolute coordinates (upper case path
            command letter)"""
            return self.command.is_absolute

        @property
        def args(self) -> List[float]:
            """Returns path command arguments as tuple of floats"""
            return self.command.args

        @property
        def control_points(self) -> List[Vector2d]:
            """Returns list of path command control points"""
            return list(
                self.command.control_points(
                    self.cfirst_point,
                    self.cprevious_end_point,
                    self.cprev2_control_point,
                )
            )

        @property
        def end_point(self) -> Vector2d:
            """Returns last control point of path command"""
            return Vector2d(self.cend_point)

        @property
        def cend_point(self) -> complex:
            """Returns last control point of path command (in complex form)"""
            return self.command.cend_point(self.cfirst_point, self.cprevious_end_point)

        def reverse(self) -> PathCommand:
            """Reverse path command"""
            return self.command.reverse(self.cend_point, self.cprevious_end_point)

        def to_curve(self) -> Curve:
            """Convert command to :py:class:`Curve`
            Curve().to_curve() returns a copy
            """
            return self.command.to_curve(
                self.cprevious_end_point, self.cprev2_control_point
            )

        def to_curves(self) -> List[Curve]:
            """Convert command to list of :py:class:`Curve` commands"""
            return self.command.to_curves(
                self.cprevious_end_point, self.cprev2_control_point
            )

        def to_absolute(self) -> AbsolutePathCommand:
            """Return relative counterpart for relative commands or copy for absolute"""
            return self.command.to_absolute(self.cprevious_end_point)

        def to_non_shorthand(self) -> AbsolutePathCommand:
            """Returns an absolute non-shorthand command

            .. versionadded:: 1.4
            """
            return self.command.to_non_shorthand(
                self.cprevious_end_point, self.cprev2_control_point
            )

        def split(self, time) -> Tuple[Path.PathCommandProxy, Path.PathCommandProxy]:
            """Split this path command into two PathCommandProxy segments.
            Raises ValueError for Move commands.

            .. versionadded:: 1.4
            """
            result = self.command.split(
                self.cfirst_point,
                self.cprevious_end_point,
                self.cprev2_control_point,
                time,
            )
            p1 = Path.PathCommandProxy(
                result[0],
                self.cfirst_point,
                self.previous_end_point,
                self.prev2_control_point,
            )
            prev2: ComplexLike = (
                0j if len(p1.control_points) < 2 else p1.control_points[-2]
            )
            p2 = Path.PathCommandProxy(
                result[1], self.cfirst_point, p1.end_point, prev2
            )
            return (p1, p2)

        def cpoint(self, time) -> complex:
            """Returns the coordinates of the Bezier curve evaluated at t as complex number.

            .. versionadded:: 1.4
            """
            return self.command.cpoint(
                self.cfirst_point,
                self.cprevious_end_point,
                self.cprev2_control_point,
                time,
            )

        def point(self, time) -> Vector2d:
            """Returns the coordinates of the Bezier curve evaluated at t as :class:`Vector2d`.

            .. versionadded:: 1.4
            """
            return self.command.point(
                self.cfirst_point,
                self.cprevious_end_point,
                self.cprev2_control_point,
                time,
            )

        def length(self, t0=0, t1=1, settings=LengthSettings()) -> float:
            """Get the length of the command between t0 and t1 in user units

            .. versionadded:: 1.4
            """
            return self.command.length(
                self.cfirst_point,
                self.cprevious_end_point,
                self.cprev2_control_point,
                t0,
                t1,
                settings,
            )

        def ilength(self, length, settings=ILengthSettings()) -> float:
            """Tries to compute the time t at which the path segment has the given
            length along its trajectory

            .. versionadded:: 1.4
            """
            return self.command.ilength(
                self.cfirst_point,
                self.cprevious_end_point,
                self.cprev2_control_point,
                length,
                settings,
            )

        def cunit_tangent(self, t) -> complex:
            """Returns the unit tangent at t as complex number

            .. versionadded:: 1.4
            """
            return self.command.cunit_tangent(
                self.cfirst_point,
                self.cprevious_end_point,
                self.cprev2_control_point,
                t,
            )

        def unit_tangent(self, t) -> Vector2d:
            """Returns the unit tangent at t as :class:`inkex.Vector2D`

            .. versionadded::  1.4
            """
            return self.command.unit_tangent(
                self.cfirst_point,
                self.cprevious_end_point,
                self.cprev2_control_point,
                t,
            )

        def __str__(self):
            return str(self.command)

        def __repr__(self):
            return "<" + self.__class__.__name__ + ">" + repr(self.command)

    def __init__(self, path_d=None) -> None:
        super().__init__()
        if isinstance(path_d, str):
            # Returns a generator returning PathCommand objects
            path_d = self.parse_string(path_d)
        elif isinstance(path_d, CubicSuperPath):
            path_d = path_d.to_path()

        for item in path_d or ():
            if isinstance(item, PathCommand):
                self.append(item)
            elif isinstance(item, (list, tuple)) and len(item) == 2:
                if isinstance(item[1], (list, tuple)):
                    self.append(PathCommand.letter_to_class(item[0])(*item[1]))
                else:
                    if len(self) == 0:
                        self.append(Move(*item))
                    else:
                        self.append(Line(*item))
            else:
                raise TypeError(
                    f"Bad path type: {type(path_d).__name__}"
                    f"({type(item).__name__}, ...): {item}"
                )

    @classmethod
    def parse_string(cls, path_d):
        """Parse a path string and generate segment objects"""
        for cmd, numbers in LEX_REX.findall(path_d):
            args = list(strargs(numbers))
            cmd = PathCommand.letter_to_class(cmd)
            i = 0
            while i < len(args) or cmd.nargs == 0:
                if len(args[i : i + cmd.nargs]) != cmd.nargs:
                    return
                seg = cmd(*args[i : i + cmd.nargs])
                i += cmd.nargs
                cmd = seg.next_command
                yield seg

    def bounding_box(self) -> Optional[BoundingBox]:
        """Return bounding box of the Path"""
        if not self:
            return None
        iterator = self.proxy_iterator()
        proxy = next(iterator)
        bbox = BoundingBox(proxy.first_point.x, proxy.first_point.y)
        try:
            while True:
                proxy = next(iterator)
                proxy.command.update_bounding_box(
                    complex(proxy.first_point),
                    [
                        proxy.cprev2_control_point,
                        proxy.cprevious_end_point,
                    ],
                    bbox,
                )
        except StopIteration:
            return bbox

    def append(self, cmd):
        """Append a command to this path."""
        try:
            cmd.letter  # pylint: disable=pointless-statement
            super().append(cmd)
        except AttributeError:
            self.extend(cmd)
            warnings.warn(
                "Passing a list to Path.add is deprecated, please use Path.extend",
                category=DeprecationWarning,
            )

    def translate(self, x, y, inplace=False):  # pylint: disable=invalid-name
        """Move all coords in this path by the given amount"""
        return self.transform(Transform(translate=(x, y)), inplace=inplace)

    def scale(self, x, y, inplace=False):  # pylint: disable=invalid-name
        """Scale all coords in this path by the given amounts"""
        return self.transform(Transform(scale=(x, y)), inplace=inplace)

    def rotate(self, deg, center=None, inplace=False):
        """Rotate the path around the given point"""
        if center is None:
            # Default center is center of bbox
            bbox = self.bounding_box()
            if bbox:
                center = bbox.center
            else:
                center = Vector2d()
        center = Vector2d(center)
        return self.transform(
            Transform(rotate=(deg, center.x, center.y)), inplace=inplace
        )

    @property
    def control_points(self) -> Iterator[Vector2d]:
        """Returns all control points of the Path"""
        prev: complex = 0
        prev_prev: complex = 0
        first: complex = 0

        seg: PathCommand
        for seg in self:
            cpts = seg.ccontrol_points(first, prev, prev_prev)
            if seg.letter in "zZmM":
                first = cpts[-1]
            for cpt in cpts:
                prev_prev = prev
                prev = cpt
                yield Vector2d(cpt)

    @property
    def cend_points(self) -> Iterator[complex]:
        """Complex version of end_points"""
        prev = 0j
        first = 0j

        seg: PathCommand
        for seg in self:
            end_point = seg.cend_point(first, prev)
            if seg.letter in "zZmM":
                first = end_point
            prev = end_point
            yield end_point

    @property
    def end_points(self) -> Iterator[Vector2d]:
        """Returns all endpoints of all path commands (i.e. the nodes)"""
        for i in self.cend_points:
            yield Vector2d(i)

    def transform(self, transform, inplace=False):
        """Convert to new path"""
        result = Path()
        previous = 0j
        previous_new = 0j
        start_zone = True
        first = 0j
        first_new = 0j

        seg: PathCommand
        for i, seg in enumerate(self):
            if start_zone:
                first = seg.cend_point(first, previous)

            if seg.letter in "hHVv":
                seg = seg.to_line(previous)

            if seg.is_relative:
                new_seg = (
                    seg.to_absolute(previous)
                    .transform(transform)
                    .to_relative(previous_new)
                )
            else:
                new_seg = seg.transform(transform)

            if start_zone:
                first_new = new_seg.cend_point(first_new, previous_new)

            if inplace:
                self[i] = new_seg
            else:
                result.append(new_seg)
            previous = seg.cend_point(first, previous)
            previous_new = new_seg.cend_point(first_new, previous_new)
            start_zone = seg.letter in "zZ"
        if inplace:
            return self
        return result

    def reverse(self):
        """Returns a reversed path"""
        result = Path()
        *_, first = self.cend_points
        closer = None

        # Go through the path in reverse order
        for index, prcom in reversed(list(enumerate(self.proxy_iterator()))):
            if prcom.letter in "MmZz":
                if closer is not None:
                    if len(result) > 0 and result[-1].letter in "LlVvHh":
                        result.pop()  # We can replace simple lines with Z
                    result.append(closer)  # replace with same type (rel or abs)
                if prcom.letter in "Zz":
                    closer = prcom.command
                else:
                    closer = None

            if index == 0:
                if prcom.letter == "M":
                    result.insert(0, Move(first))
                elif prcom.letter == "m":
                    result.insert(0, move(first))
            else:
                result.append(prcom.reverse())

        return result

    def break_apart(self) -> List[Path]:
        """Breaks apart a path into its subpaths

        .. versionadded:: 1.3
        """
        result = [Path()]
        current = result[0]

        for cmnd in self.proxy_iterator():
            if cmnd.letter.lower() == "m":
                current = Path()
                result.append(current)
                current.append(Move(cmnd.cend_point))
            else:
                current.append(cmnd.command)
        # Remove all subpaths that are empty or only contain move commands
        return [
            i
            for i in result
            if len(i) != 0 and not all(j.letter.lower() == "m" for j in i)
        ]

    def close(self):
        """Attempt to close the last path segment"""
        if self and not self[-1].letter in "zZ":
            self.append(ZoneClose())

    def proxy_iterator(self) -> Iterator[PathCommandProxy]:
        """
        Yields :py:class:`AugmentedPathIterator`

        :rtype: Iterator[ Path.PathCommandProxy ]
        """

        previous = 0j
        prev_prev = 0j
        first = 0j
        seg: PathCommand
        for seg in self:
            if seg.letter in "zZmM":
                first = seg.cend_point(first, previous)
            yield Path.PathCommandProxy(seg, first, previous, prev_prev)
            if seg.letter in "ctqsCTQS":
                prev_prev = seg.ccontrol_points(first, previous, prev_prev)[-2]
            previous = seg.cend_point(first, previous)

    def subpath_iterator(self):
        """Yield Path for each subpath."""
        start_id = 0

        for i, seg in enumerate(self):
            if isinstance(seg, (move, Move)):
                if start_id > -1 and i > 0:  # add previous path (open path)
                    yield Path(self[start_id:i])
                start_id = i
            elif isinstance(seg, (zoneClose, ZoneClose)):  # add current path (closed)
                yield Path(self[start_id : i + 1])
                start_id = -1
            elif i == len(self) - 1 and start_id > -1:  # add last path (open)
                yield Path(self[start_id:])

    def to_absolute(self):
        """Convert this path to use only absolute coordinates"""
        return self._to_absolute(True)

    def to_non_shorthand(self) -> Path:
        """Convert this path to use only absolute non-shorthand coordinates

        .. versionadded:: 1.1
        """
        return self._to_absolute(False)

    def _to_absolute(self, shorthand: bool) -> Path:
        """Make entire Path absolute.

        Args:
            shorthand (bool): If false, then convert all shorthand commands to
                non-shorthand.

        Returns:
            Path: the input path, converted to absolute coordinates.
        """

        abspath = Path()

        previous = 0j
        first = 0j
        seg: PathCommand
        for seg in self:
            if seg.letter in "mM":
                first = seg.cend_point(first, previous)

            if shorthand:
                abspath.append(seg.to_absolute(previous))
            else:
                if abspath and abspath[-1].letter in "QC":
                    prev_control = list(abspath[-1].control_points(0, 0, 0))[-2]
                else:
                    prev_control = previous

                abspath.append(seg.to_non_shorthand(previous, prev_control))

            previous = seg.cend_point(first, previous)

        return abspath

    def to_relative(self):
        """Convert this path to use only relative coordinates"""
        abspath = Path()

        previous = 0j
        first = 0j
        seg: PathCommand
        for seg in self:
            if seg.letter in "mM":
                first = seg.cend_point(first, previous)

            abspath.append(seg.to_relative(previous))
            previous = seg.cend_point(first, previous)

        return abspath

    def __str__(self):
        return " ".join([str(seg) for seg in self])

    def __add__(self, other):
        acopy = copy.deepcopy(self)
        if isinstance(other, str):
            other = Path(other)
        if isinstance(other, list):
            acopy.extend(other)
        return acopy

    def to_arrays(self):
        """Returns path in format of parsePath output, returning arrays of absolute
        command data

        .. deprecated:: 1.0
            This is compatibility function for older API. Should not be used in new code

        """
        return [[seg.letter, list(seg.args)] for seg in self.to_non_shorthand()]

    def to_superpath(self):
        """Convert this path into a cubic super path"""
        return CubicSuperPath(self)

    def copy(self):
        """Make a copy"""
        return copy.deepcopy(self)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        if self.callback is not None:
            self.callback(self)  # pylint: disable=not-callable


class CubicSuperPath(list):
    """
    A conversion of a path into a predictable list of cubic curves which
    can be operated on as a list of simplified instructions.

    When converting back into a path, all lines, arcs etc will be converted
    to curve instructions.

    Structure is held as [SubPath[(point_a, bezier, point_b), ...], ...]
    """

    def __init__(self, items):
        super().__init__()
        self._closed = True
        self._prev = 0j
        self._prev_prev = 0j

        if isinstance(items, str):
            items = Path(items)

        if isinstance(items, Path):
            for item in items:
                self.append_path_command(item)
            return

        for item in items:
            self.append(item)

    def __str__(self):
        return str(self.to_path())

    def append_node_with_handles(self, command: List[Tuple[float, float]]):
        """First item: left handle, second item: node coords,
        third item: right handle"""
        if self._closed:
            # Closed means that the previous segment is closed so we need a new one
            # We always append to the last open segment. CSP starts out closed.
            self._closed = False
            super().append([])

        self[-1].append(command)
        self._prev_prev = command[0][0] + command[0][1] * 1j
        self._prev = command[1][0] + command[1][1] * 1j

    def append_path_command(self, command: PathCommand):
        """Append a path command.

        For ordinary commands:

        * old last entry -> [[.., ..], [.., ..], [x1, y1]]
        * new last entry -> [[x2, y2], [x3, y3], [x3, y3]]

        The last tuple is duplicated (retracted handle): either it's the last command
        of the subpath, then the handle will stay retracted, or it will be replaced
        with the next path command.
        """
        if command.letter in "mM":
            carg = command.cend_point(self._first, self._prev)
            arg = Vector2d.c2t(carg)
            super().append([[arg[:], arg[:], arg[:]]])
            self._prev = self._prev_prev = carg
            self._closed = False
            return
        if command.letter in "zZ" and self:
            # This duplicates the first segment to 'close' the path
            self[-1].append([self[-1][0][0][:], self[-1][0][1][:], self[-1][0][2][:]])
            # Then adds a new subpath for the next shape (if any)
            # self._closed = True
            self._prev = self._first
            return
        if command.letter in "aA":
            # Arcs are made up of (possibly) more than one curve, depending on their
            # angle (approximated)
            for arc_curve in command.to_curves(self._prev, self._prev_prev):
                self.append_path_command(arc_curve)
            return
        # Handle regular curves.

        if self._closed:
            # Previous segment is closed. Append a new segment first.
            self._closed = False
            super().append([])

        cp1, cp2, cp3 = command.ccurve_points(0j, self._prev, self._prev_prev)

        item = [Vector2d.c2t(cp1), Vector2d.c2t(cp2), Vector2d.c2t(cp3)]
        self._prev = cp3
        if not command.letter in "QT":
            self._prev_prev = cp2
        else:
            self._prev_prev = command.ccontrol_points(0j, self._prev, self._prev_prev)[
                0
            ]

        if self[-1]:  # There exists a previous segment, replace its outgoing handle.
            self[-1][-1][-1] = item[0]
        # Append the segment with the last coordinate (node pos) repeated.
        self[-1].append(item[1:] + [item[-1][:]])

    def append(self, item):
        """Append a segment/node to the superpath and update the internal state.

        item may be specified in any of the following formats:

        * ``PathCommand``
        * ``[str, List[float]]`` - A path command letter and its arguments
        * ``[[float, float], [float, float], [float, float]]`` - Incoming handle, node,
           outgoing handle.
        * ``List[[float, float], [float, float], [float, float]]`` - An entire subpath.

        """
        if isinstance(item, list) and len(item) == 2 and isinstance(item[0], str):
            item = PathCommand.letter_to_class(item[0])(*item[1])
        if isinstance(item, PathCommand):
            self.append_path_command(item)
            return

        if isinstance(item, list):
            # Item is a subpath: List[Handle, node, Handle]. Just append the
            # subpath, and update the prev/ prev_prev positions.
            if (
                (len(item) != 3 or not all(len(bit) == 2 for bit in item))
                and len(item[0]) == 3
                and all(len(bit) == 2 for bit in item[0])
            ):
                super().append(self._clean(item))

            elif len(item) == 3 and all(len(bit) == 2 for bit in item):
                # Item is already a csp segment [Handle, node, Handle].
                if self._closed:
                    # Closed means that the previous segment is closed so we need a new
                    # one.
                    # We always append to the last open segment. CSP starts out closed.
                    self._closed = False
                    super().append([])

                # Item is already a csp segment and has already been shifted.
                self[-1].append([i.copy() for i in item])
            else:
                raise ValueError(f"Unknown super curve list format: {item}")

            self._prev_prev = Vector2d.t2c(self[-1][-1][0])
            self._prev = Vector2d.t2c(self[-1][-1][1])
        else:
            raise ValueError(f"Unknown super curve list format: {item}")

    def _clean(self, lst):
        """Recursively clean lists so they have the same type"""
        if isinstance(lst, (tuple, list)):
            return [self._clean(child) for child in lst]
        return lst

    @property
    def _first(self):
        try:
            return self[-1][0][0][0] + self[-1][0][0][1] * 1j
        except IndexError:
            return 0 + 0j

    def to_path(self, curves_only=False, rtol=1e-5, atol=1e-8):
        """Convert the super path back to an svg path

        Arguments: see :func:`to_segments` for parameters
        """
        return Path(list(self.to_segments(curves_only, rtol, atol)))

    def to_segments(self, curves_only=False, rtol=1e-5, atol=1e-8):
        """Generate a set of segments for this cubic super path

        Arguments:
            curves_only (bool, optional): If False, curves that can be represented
                by Lineto / ZoneClose commands, will be. Defaults to False.
            rtol (float, optional): relative tolerance, passed to :func:`is_line` and
                :func:`inkex.transforms.ImmutableVector2d.is_close` for checking if a
                line can be replaced by a ZoneClose command. Defaults to 1e-5.

                .. versionadded:: 1.2
            atol: absolute tolerance, passed to :func:`is_line` and
                :func:`inkex.transforms.ImmutableVector2d.is_close`. Defaults to 1e-8.

                .. versionadded:: 1.2
        """
        for subpath in self:
            previous = []
            for segment in subpath:
                if not previous:
                    yield Move(Vector2d(segment[1]))
                elif self.is_line(previous, segment, rtol, atol) and not curves_only:
                    if segment is subpath[-1] and Vector2d(segment[1]).is_close(
                        Vector2d(subpath[0][1]), rtol, atol
                    ):
                        yield ZoneClose()
                    else:
                        yield Line(Vector2d(segment[1]))
                else:
                    yield Curve(
                        Vector2d(previous[2]),
                        Vector2d(segment[0]),
                        Vector2d(segment[1]),
                    )
                previous = segment

    def transform(self, transform):
        """Apply a transformation matrix to this super path"""
        return self.to_path().transform(transform).to_superpath()

    @staticmethod
    def is_on(pt_a, pt_b, pt_c, tol=1e-8):
        """Checks if point pt_a is on the line between points pt_b and pt_c

        .. versionadded:: 1.2
        """
        return CubicSuperPath.collinear(pt_a, pt_b, pt_c, tol) and (
            CubicSuperPath.within(pt_a[0], pt_b[0], pt_c[0])
            if abs(pt_a[0] - pt_b[0]) > 1e-13
            else CubicSuperPath.within(pt_a[1], pt_b[1], pt_c[1])
        )

    @staticmethod
    def collinear(pt_a, pt_b, pt_c, tol=1e-8):
        """Checks if points pt_a, pt_b, pt_c lie on the same line,
        i.e. that the cross product (b-a) x (c-a) < tol

        .. versionadded:: 1.2
        """
        return (
            abs(
                (pt_b[0] - pt_a[0]) * (pt_c[1] - pt_a[1])
                - (pt_c[0] - pt_a[0]) * (pt_b[1] - pt_a[1])
            )
            < tol
        )

    @staticmethod
    def within(val_b, val_a, val_c):
        """Checks if float val_b is between val_a and val_c

        .. versionadded:: 1.2
        """
        return val_a <= val_b <= val_c or val_c <= val_b <= val_a

    @staticmethod
    def is_line(previous, segment, rtol=1e-5, atol=1e-8):
        """Check whether csp segment (two points) can be expressed as a line has
        retracted handles or the handles can be retracted without loss of information
        (i.e. both handles lie on the line)

        .. versionchanged:: 1.2
            Previously, it was only checked if both control points have retracted
            handles. Now it is also checked if the handles can be retracted without
            (visible) loss of information (i.e. both handles lie on the line connecting
            the nodes).

        Arguments:
            previous: first node in superpath notation
            segment: second node in superpath notation
            rtol (float, optional): relative tolerance, passed to
                :func:`inkex.transforms.ImmutableVector2d.is_close` for checking handle
                retraction. Defaults to 1e-5.

                .. versionadded:: 1.2
            atol (float, optional): absolute tolerance, passed to
                :func:`inkex.transforms.ImmutableVector2d.is_close` for checking handle
                retraction and
                :func:`inkex.paths.CubicSuperPath.is_on` for checking if all points
                (nodes + handles) lie on a line. Defaults to 1e-8.

                .. versionadded:: 1.2
        """

        retracted = isclose(
            Vector2d(previous[1]), Vector2d(previous[2]), rel_tol=rtol, abs_tol=atol
        ) and isclose(
            Vector2d(segment[0]), Vector2d(segment[1]), rel_tol=rtol, abs_tol=atol
        )

        if retracted:
            return True

        # Can both handles be retracted without loss of information?
        # Definitely the case if the handles lie on the same line as the two nodes and
        # in the correct order
        # E.g. cspbezsplitatlength outputs non-retracted handles when splitting a
        # straight line
        return CubicSuperPath.is_on(
            segment[0], segment[1], previous[2], atol
        ) and CubicSuperPath.is_on(previous[2], previous[1], segment[0], atol)



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2005 Aaron Spike, aaron@ekips.org
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
Perspective approach & math by Dmitry Platonov, shadowjack@mail.ru, 2006
"""

import inkex
from inkex.localization import inkex_gettext as _

X, Y = range(2)

try:
    import numpy as np
    import numpy.linalg as lin

    FLOAT = np.float64
except ImportError:
    np = None


class Perspective(inkex.EffectExtension):
    """Apply a perspective to a path/group of paths"""

    def effect(self):
        if np is None:
            raise inkex.AbortExtension(
                _(
                    "Failed to import the numpy or numpy.linalg modules. "
                    "These modules are required by this extension. Please install them."
                    "  On a Debian-like system this can be done with the command, "
                    "sudo apt-get install python-numpy."
                )
            )
        if len(self.svg.selection) != 2:
            raise inkex.AbortExtension(
                _("This extension requires two selected objects.")
            )

        obj, envelope = self.svg.selection

        if isinstance(obj, (inkex.PathElement, inkex.Group)):
            if isinstance(envelope, inkex.PathElement):
                path = envelope.path.transform(
                    envelope.composed_transform()
                ).to_superpath()

                if len(path) < 1 or len(path[0]) < 4:
                    raise inkex.AbortExtension(
                        _(
                            "This extension requires that the second path be four "
                            "nodes long."
                        )
                    )

                dip = np.zeros((4, 2), dtype=FLOAT)
                for i in range(4):
                    dip[i][0] = path[0][i][1][0]
                    dip[i][1] = path[0][i][1][1]

                # Get bounding box plus any extra composed transform of parents.
                bbox = obj.bounding_box(obj.getparent().composed_transform())

                sip = np.array(
                    [
                        [bbox.left, bbox.bottom],
                        [bbox.left, bbox.top],
                        [bbox.right, bbox.top],
                        [bbox.right, bbox.bottom],
                    ],
                    dtype=FLOAT,
                )
            else:
                if isinstance(envelope, inkex.Group):
                    raise inkex.AbortExtension(
                        _(
                            "The second selected object is a group, not a"
                            " path.\nTry using Object->Ungroup."
                        )
                    )
                raise inkex.AbortExtension(
                    _(
                        "The second selected object is not a path.\nTry using"
                        " the procedure Path->Object to Path."
                    )
                )
        else:
            raise inkex.AbortExtension(
                _(
                    "The first selected object is neither a path nor a group.\nTry "
                    "using the procedure Path->Object to Path."
                )
            )

        solmatrix = np.zeros((8, 8), dtype=FLOAT)
        free_term = np.zeros(8, dtype=FLOAT)
        for i in (0, 1, 2, 3):
            solmatrix[i][0] = sip[i][0]
            solmatrix[i][1] = sip[i][1]
            solmatrix[i][2] = 1
            solmatrix[i][6] = -dip[i][0] * sip[i][0]
            solmatrix[i][7] = -dip[i][0] * sip[i][1]
            solmatrix[i + 4][3] = sip[i][0]
            solmatrix[i + 4][4] = sip[i][1]
            solmatrix[i + 4][5] = 1
            solmatrix[i + 4][6] = -dip[i][1] * sip[i][0]
            solmatrix[i + 4][7] = -dip[i][1] * sip[i][1]
            free_term[i] = dip[i][0]
            free_term[i + 4] = dip[i][1]

        res = lin.solve(solmatrix, free_term)
        projmatrix = np.array(
            [[res[0], res[1], res[2]], [res[3], res[4], res[5]], [res[6], res[7], 1.0]],
            dtype=FLOAT,
        )

        self.process_object(obj, projmatrix)

    def process_object(self, obj, matrix):
        if isinstance(obj, inkex.PathElement):
            self.process_path(obj, matrix)
        elif isinstance(obj, inkex.Group):
            self.process_group(obj, matrix)

    def process_group(self, group, matrix):
        """Go through all groups to process all paths inside them"""
        for node in group:
            self.process_object(node, matrix)

    def process_path(self, element, matrix):
        """Apply the transformation to the selected path"""
        point = (
            element.path.to_absolute()
            .transform(element.composed_transform())
            .to_superpath()
        )
        for subs in point:
            for csp in subs:
                csp[0] = self.project_point(csp[0], matrix)
                csp[1] = self.project_point(csp[1], matrix)
                csp[2] = self.project_point(csp[2], matrix)
        element.path = inkex.Path(point).transform(-element.composed_transform())

    @staticmethod
    def project_point(point, matrix):
        """Apply the matrix to the given point"""
        return [
            (point[X] * matrix[0][0] + point[Y] * matrix[0][1] + matrix[0][2])
            / (point[X] * matrix[2][0] + point[Y] * matrix[2][1] + matrix[2][2]),
            (point[X] * matrix[1][0] + point[Y] * matrix[1][1] + matrix[1][2])
            / (point[X] * matrix[2][0] + point[Y] * matrix[2][1] + matrix[2][2]),
        ]


if __name__ == "__main__":
    Perspective().run()



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2005 Aaron Spike, aaron@ekips.org
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import random
import inkex
from inkex import turtle as pturtle


class Lindenmayer(inkex.GenerateExtension):
    def add_arguments(self, pars):
        pars.add_argument("--tab")
        pars.add_argument("--order", type=int, default=3, help="number of iteration")
        pars.add_argument(
            "--langle", type=float, default=16.0, help="angle for turning left"
        )
        pars.add_argument(
            "--rangle", type=float, default=16.0, help="angle for turning right"
        )
        pars.add_argument("--step", type=float, default=25.0, help="step size")
        pars.add_argument(
            "--randomizestep", type=float, default=0.0, help="randomize step"
        )
        pars.add_argument(
            "--randomizeangle", type=float, default=0.0, help="randomize angle"
        )
        pars.add_argument("--axiom", default="++F", help="initial state of system")
        pars.add_argument(
            "--rules", default="F=FF-[-F+F+F]+[+F-F-F]", help="replacement rules"
        )
        self.stack = []
        self.turtle = pturtle.pTurtle()

    def iterate(self):
        self.rules = dict(
            [
                map((lambda s: s.strip()), i.split("="))
                for i in self.options.rules.upper().split(";")
                if i.count("=") == 1
            ]
        )
        string = self.__recurse(self.options.axiom.upper(), 0)
        self.__compose_path(string)
        return self.turtle.getPath()

    def __compose_path(self, string):
        self.turtle.pu()
        point = self.svg.namedview.center
        self.turtle.setpos(point)
        self.turtle.pd()
        for c in string:
            if c in "ABCDEF":
                self.turtle.pd()
                self.turtle.fd(
                    self.options.step
                    * (random.normalvariate(1.0, 0.01 * self.options.randomizestep))
                )
            elif c in "GHIJKL":
                self.turtle.pu()
                self.turtle.fd(
                    self.options.step
                    * (random.normalvariate(1.0, 0.01 * self.options.randomizestep))
                )
            elif c == "+":
                self.turtle.lt(
                    self.options.langle
                    * (random.normalvariate(1.0, 0.01 * self.options.randomizeangle))
                )
            elif c == "-":
                self.turtle.rt(
                    self.options.rangle
                    * (random.normalvariate(1.0, 0.01 * self.options.randomizeangle))
                )
            elif c == "|":
                self.turtle.lt(180)
            elif c == "[":
                self.stack.append([self.turtle.getpos(), self.turtle.getheading()])
            elif c == "]":
                self.turtle.pu()
                pos, heading = self.stack.pop()
                self.turtle.setpos(pos)
                self.turtle.setheading(heading)

    def __recurse(self, rule, level):
        level_string = ""
        for c in rule:
            if level < self.options.order:
                try:
                    level_string = level_string + self.__recurse(
                        self.rules[c], level + 1
                    )
                except KeyError:
                    level_string = level_string + c
            else:
                level_string = level_string + c
        return level_string

    def generate(self):
        self.options.step = self.svg.unittouu(str(self.options.step) + "px")
        sty = {
            "stroke-linejoin": "miter",
            "stroke-width": str(self.svg.unittouu("1px")),
            "stroke-opacity": "1.0",
            "fill-opacity": "1.0",
            "stroke": "#000000",
            "stroke-linecap": "butt",
            "fill": "none",
        }
        return inkex.PathElement(style=str(inkex.Style(sty)), d=self.iterate())


if __name__ == "__main__":
    Lindenmayer().run()



#!/usr/bin/env python3

# Copyright (C) 2024 Jonathan Neuhauser <jonathan.neuhauser@outlook.com>
# SPDX-License-Identifier: GPL-2.0-or-later

"""
Common stream reading functions
"""

from io import BufferedIOBase, BytesIO
import struct
from typing import Union


class ReaderWrapper:
    def __init__(self, stream: BufferedIOBase):
        self.stream = stream

    def read(self, num: int) -> bytes:
        read = self.stream.read(num)
        assert len(read) == num, "Unexpected EOF"
        return read

    def tell(self) -> int:
        return self.stream.tell()


class BitReader(ReaderWrapper):
    def __init__(self, stream: BufferedIOBase):
        super().__init__(stream)
        self.current_byte = 0
        self.bit_shift = 8

    def read_bit(self) -> bool:
        if self.bit_shift == 8:
            byte = self.stream.read(1)
            if not byte:  # End of stream
                raise EOFError("No more bits available")
            self.current_byte = ord(byte)
            self.bit_shift = 0
        result = bool((self.current_byte >> self.bit_shift) & 1)
        self.bit_shift += 1
        return result

    def clear_bits(self):
        self.bit_shift = 8


class BinaryParser:
    def __init__(self, data: Union[ReaderWrapper, bytes, BufferedIOBase]):
        if isinstance(data, bytes):
            data = BytesIO(data)
        if isinstance(data, BufferedIOBase):
            data = ReaderWrapper(data)
        self.data = data
        assert self.data is not None, f"Bad type, {type(data)}"

    def read(self, num_bytes: int) -> bytes:
        return self.data.read(num_bytes)

    def get_pos(self):
        return self.data.tell()

    def read_u8(self) -> int:
        return int.from_bytes(self.read(1), byteorder="little", signed=False)

    def read_u16(self) -> int:
        return int.from_bytes(self.read(2), byteorder="little", signed=False)

    def read_u32(self) -> int:
        return int.from_bytes(self.read(4), byteorder="little", signed=False)

    def read_u64(self) -> int:
        return int.from_bytes(self.read(8), byteorder="little", signed=False)

    def read_i32(self) -> int:
        return int.from_bytes(self.read(4), byteorder="little", signed=True)

    def read_float(self) -> float:
        return struct.unpack("f", self.read(4))[0]

    def read_double(self) -> float:
        return struct.unpack("d", self.read(8))[0]

    def read_tag(self) -> str:
        tag = self.read(4).decode("latin-1")[::-1]
        return tag



# coding=utf-8
#
# Copyright (C) 2007 Martin Owens
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
#
"""
Python barcode renderer for Code93 barcodes. Designed for use with Inkscape.
"""

from .Base import Barcode

PALLET = list("0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ-. $/+%")
PALLET.append("($)")
PALLET.append("(/)")
PALLET.append("(+)")
PALLET.append("(%)")
PALLET.append("MARKER")

MAP = dict((PALLET[i], i) for i in range(len(PALLET)))


def get_map(array):
    """Extended ENCODE maps for full ASCII Code93"""
    result = {}
    pos = 10
    for char in array:
        result[chr(char)] = PALLET[pos]
        pos += 1
    return result


# MapA is eclectic, but B, C, D are all ASCII ranges
MAP_A = get_map(
    [
        27,
        28,
        29,
        30,
        31,
        59,
        60,
        61,
        62,
        63,
        91,
        92,
        93,
        94,
        95,
        123,
        124,
        125,
        126,
        127,
        0,
        64,
        96,
        127,
        127,
        127,
    ]
)  # %
MAP_B = get_map(range(1, 26))  # $
MAP_C = get_map(range(33, 58))  # /
MAP_D = get_map(range(97, 122))  # +

ENCODE = [
    "100010100",
    "101001000",
    "101000100",
    "101000010",
    "100101000",
    "100100100",
    "100100010",
    "101010000",
    "100010010",
    "100001010",
    "110101000",
    "110100100",
    "110100010",
    "110010100",
    "110010010",
    "110001010",
    "101101000",
    "101100100",
    "101100010",
    "100110100",
    "100011010",
    "101011000",
    "101001100",
    "101000110",
    "100101100",
    "100010110",
    "110110100",
    "110110010",
    "110101100",
    "110100110",
    "110010110",
    "110011010",
    "101101100",
    "101100110",
    "100110110",
    "100111010",
    "100101110",
    "111010100",
    "111010010",
    "111001010",
    "101101110",
    "101110110",
    "110101110",
    "100100110",
    "111011010",
    "111010110",
    "100110010",
    "101011110",
    "",
]


class Code93(Barcode):
    """Encode Code93 Barcode"""

    def encode(self, text):
        # start marker
        bits = ENCODE[MAP.get("MARKER", -1)]

        # Extend to ASCII charset ( return Array )
        text = Code93.encode_ascii(text)

        # Calculate the checksums
        text.append(Code93.checksum(text, 20))  # C
        text.append(Code93.checksum(text, 15))  # K

        # Now convert text into the ENCODE bits (black and white stripes)
        for char in text:
            bits = bits + ENCODE[MAP.get(char, -1)]

        # end marker and termination bar
        return bits + ENCODE[MAP.get("MARKER", -1)] + "1"

    @staticmethod
    def checksum(text, mod):
        """Generate a code 93 checksum"""
        weight = len(text) % mod
        check = 0
        for char in text:
            check = check + (MAP[char] * weight)
            # Reset the weight is required
            weight -= 1
            if weight == 0:
                weight = mod

        return PALLET[check % 47]

    @staticmethod
    def encode_ascii(text):
        """Some characters need re-ENCODE into the code93 specification"""
        result = []
        for char in text:
            if char in MAP:
                result.append(char)
            elif char in MAP_A:
                result.append("(%)")
                result.append(MAP_A[char])
            elif char in MAP_B:
                result.append("($)")
                result.append(MAP_B[char])
            elif char in MAP_C:
                result.append("(/)")
                result.append(MAP_C[char])
            elif char in MAP_D:
                result.append("(+)")
                result.append(MAP_D[char])
        return result



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2020-2022 Jonathan Neuhauser <jonathan.neuhauser@outlook.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Global export settings / object storage"""


from typing import Dict, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from ..xamlobjects import ResourceDictionary


class XAMLExportGlobalSettings:
    """Storage for XAML export converter settings"""

    def __init__(self) -> None:
        self.target = "wpf"
        self.mode = "lowlevel"
        self.indent = 4
        self.swatches = "color"
        self.referencing_type = "DrawingImage"
        self.xaml_namespaces: Dict[Optional[str], str] = {}
        self.xaml_namespaces_reversed: Dict[str, Optional[str]] = {}
        self.layers_as_resources = True
        self.TOPLEVEL_DICT: "ResourceDictionary"

    @property
    def avalonia(self):
        """Returns if the settings target Avalonia"""
        return self.target != "wpf"

    @property
    def wpf(self):
        """Returns if the settings target WPF"""
        return self.target == "wpf"


GLOBAL_SETTINGS = XAMLExportGlobalSettings()



"""
VI decoders

converts Vectornator JSON data to inkvn (Vectornator mode).
This is used for Linearity Curve 5.0.x import as well (format 19).

Needs more Vectornator files for reference
"""

import base64
import logging
from typing import Any, Dict, List, Union

import inkex

import inkvn.reader.extract as ext
import inkvn.reader.text as t
from inkvn.utils import NSKeyedUnarchiver

from ..elements.artboard import VNArtboard, VNLayer, Frame
from ..elements.base import VNBaseElement, VNTransform
from ..elements.guide import VNGuideElement
from ..elements.group import VNGroupElement
from ..elements.image import VNImageElement
from ..elements.path import VNPathElement, pathGeometry
from ..elements.text import VNTextElement, singleStyledText
from ..elements.styles import (
    VNColor,
    VNGradient,
    pathStrokeStyle,
    basicStrokeStyle,
    styledElementData,
)


def read_vn_artboard(archive: Any, gid_json: Dict) -> VNArtboard:
    """
    Reads gid.json(artboard) and returns artboard class.

    Argument `archive` is needed for image embedding.
    """
    # Layers
    layers = gid_json["layers"]
    layer_list: List[VNLayer] = []
    for layer in layers:
        if layer is not None:
            layer_list.append(read_vn_layer(archive, layer))

    # Guides
    guides = gid_json["guideLayer"]["elements"]
    guide_list: List[VNGuideElement] = []
    for guide in guides:
        if guide is not None:
            guide_list.append(read_vn_element(archive, guide))

    # Background Color
    fill_color = None
    fill_gradient = None
    fill_result = read_vn_fill(gid_json)
    if isinstance(fill_result, VNGradient):
        fill_gradient = fill_result
    elif isinstance(fill_result, VNColor):
        fill_color = fill_result

    return VNArtboard(
        title=gid_json["title"],
        frame=Frame(**gid_json["frame"]),
        layers=layer_list,
        guides=guide_list,
        fillColor=fill_color,
        fillGradient=fill_gradient,
    )


def read_vn_layer(archive: Any, layer: Dict) -> VNLayer:
    """
    Read specified layer and return their attributes as class.
    """
    # children
    elements = layer["elements"]
    element_list: List[VNBaseElement] = []
    for element in elements:
        if element is not None:
            element_list.append(read_vn_element(archive, element))

    # properties
    if layer.get("properties") is not None:
        properties = layer["properties"]

    # properties did not exist in 4.9.0, format 7.
    else:
        properties = layer

    return VNLayer(
        name=properties.get("name", "Unnamed Layer"),
        opacity=properties.get("opacity", 1.0),
        isVisible=properties.get("isVisible", True),
        isLocked=properties.get("isLocked", False),
        isExpanded=properties.get("isExpanded", False),
        elements=element_list,
    )


def read_vn_element(archive: Any, element: Dict) -> VNBaseElement:
    """Traverse specified element and extract their attributes."""
    base_element_data = {
        "name": element.get("name", "Unnamed Element"),
        "blur": element.get("blur", 0.0),
        "opacity": element.get("opacity", 1.0),
        "blendMode": element.get("blendMode", 0),
        "isHidden": element.get("isHidden", False),
        "isLocked": element.get("isLocked", False),
        "localTransform": None,
    }

    try:
        # localTransform (BaseElement)
        local_transform = element.get("localTransform")
        if local_transform is not None:
            base_element_data["localTransform"] = VNTransform(**local_transform)

        # Guide (GuideElement)
        guide = element.get("subElement", {}).get("guideLine", {}).get("_0")
        if guide is not None:
            return VNGuideElement(**guide, **base_element_data)

        # Group (GroupElement)
        group = element.get("subElement", {}).get("group", {}).get("_0")
        if group is not None:
            return read_vn_group(archive, group, base_element_data)

        # Image (ImageElement)
        image = element.get("subElement", {}).get("image", {}).get("_0")
        if image is not None:
            return read_vn_image(archive, image, base_element_data)

        # Stylable (either PathElement or TextElement)
        stylable = element.get("subElement", {}).get("stylable", {}).get("_0")
        if stylable is not None:
            # clipping mask
            mask = stylable.get("mask", 0)

            # Stroke Style
            stroke_style = read_vn_stroke(stylable)

            # fill
            fill_color = None
            fill_gradient = None
            fill_result = read_vn_fill(stylable)
            if isinstance(fill_result, VNGradient):
                fill_gradient = fill_result
            elif isinstance(fill_result, VNColor):
                fill_color = fill_result

            # singleStyles (Vectornator 4.13.6, format 19)
            abstract_path = None
            single_style = (
                stylable.get("subElement", {}).get("singleStyle", {}).get("_0")
            )
            if single_style is not None:
                fill_result = read_vn_fill(stylable, single_style)
                if isinstance(fill_result, VNGradient):
                    fill_gradient = fill_result
                elif isinstance(fill_result, VNColor):
                    fill_color = fill_result

                # use single_style as abstract path
                abstract_path = single_style.get("subElement")

            # Abstract Path (PathElement)
            sub_element = stylable.get("subElement", {})
            if "abstractPath" in sub_element:
                abstract_path = sub_element["abstractPath"].get("_0")
            if abstract_path is not None:
                path_element_data = styledElementData(
                    styled_data=abstract_path,
                    mask=mask,
                    stroke=stroke_style,
                    color=fill_color,
                    grad=fill_gradient,
                )
                return read_vn_abst_path(path_element_data, base_element_data)

            # Abstract Text (TextElement)
            text_data = stylable.get("subElement", {}).get("text", {}).get("_0")
            if text_data is not None:
                text_element_data = styledElementData(
                    styled_data=text_data,
                    mask=mask,
                    stroke=stroke_style,
                    color=fill_color,
                    grad=fill_gradient,
                )
                return read_vn_abst_text(text_element_data, base_element_data)

        # if the element is unknown type:
        raise NotImplementedError(
            f"{base_element_data['name']}: This element has unknown type."
        )

    except Exception as e:
        # if the element is unknown type or an error occurred:
        logging.error(f"Error reading element: {e}", exc_info=True)
        return VNBaseElement(**base_element_data)


def read_vn_group(archive: Any, group: Dict, base_element: Dict) -> VNGroupElement:
    """Reads elements inside group and returns as VNGroupElement."""
    # get elements inside group
    group_elements = group["elements"]
    group_element_list: List[VNBaseElement] = []

    for group_element in group_elements:
        if group_element is not None:
            # get group elements recursively
            group_element_list.append(read_vn_element(archive, group_element))

    return VNGroupElement(groupElements=group_element_list, **base_element)


def read_vn_image(archive: Any, image: Dict, base_element: Dict) -> VNImageElement:
    """Reads image element, encodes image in Base64 and returns VNImageElement."""
    # relativePath contains *.dat (bitmap data)
    transform = image["transform"]
    image_file = image["imageData"]["relativePath"]
    image_data = ext.read_dat_from_zip(archive, image_file)

    # cropping
    crop_rect = None
    if image.get("cropRect") is not None:
        crop_rect = tuple(map(tuple, image.get("cropRect")))
    return VNImageElement(
        imageData=image_data, transform=transform, cropRect=crop_rect, **base_element
    )


def read_vn_abst_path(
    path_element: styledElementData, base_element: Dict
) -> VNPathElement:
    """Reads path element and returns VNPathElement."""

    def _add_path(path_data: Dict, path_geometry_list: List[pathGeometry]) -> None:
        """appends path data to list"""
        geometry = path_data.get("geometry")
        # AbstractPath
        if path_data.get("nodes") is not None:
            path_geometry_list.append(
                pathGeometry(closed=path_data["closed"], nodes=path_data["nodes"])
            )
        # SingleStyle
        elif geometry is not None:
            path_geometry_list.append(
                pathGeometry(closed=geometry["closed"], nodes=geometry["nodes"])
            )

    # Path
    path_data = (
        path_element.styled_data.get("subElement", {}).get("pathData", {}).get("_0")
    )
    path_geometry_list: List[pathGeometry] = []

    if path_data is not None:
        text_path = path_data.get("subElement", {}).get("textPath", {}).get("_0")
        if text_path is not None:
            inkex.utils.debug(f"{base_element['name']}: textOnPath is not supported.")
        _add_path(path_data, path_geometry_list)

    # compoundPath
    compound_path_data = (
        path_element.styled_data.get("subElement", {})
        .get("compoundPathData", {})
        .get("_0")
    )
    if compound_path_data is not None:
        # Path Geometries (subpath)
        subpaths = compound_path_data.get("subpaths")
        if subpaths is not None:
            for sub_element in subpaths:
                sub_stylable = (
                    sub_element.get("subElement", {}).get("stylable", {}).get("_0")
                )
                # Vectornator 4.13.5, format 16
                if sub_stylable is not None:
                    sub_path = sub_stylable["subElement"]["abstractPath"]["_0"][
                        "subElement"
                    ]["pathData"]["_0"]

                # else (Vectornator 4.13.6, format 19)
                else:
                    sub_path = sub_element

                _add_path(sub_path, path_geometry_list)

    return VNPathElement(
        mask=path_element.mask,
        fillColor=path_element.color,
        fillGradient=path_element.grad,
        strokeStyle=path_element.stroke,
        pathGeometries=path_geometry_list,
        **base_element,
    )


def read_vn_abst_text(
    text_element: styledElementData, base_element: Dict
) -> VNBaseElement:
    """
    Reads legacy text element and returns VNTextElement.
    """
    text_data = text_element.styled_data

    # will be used to return TextElement
    text_property = None
    styled_text = None

    # I cannot replicate textProperty in legacy format
    transform = text_data.get("transform")  # matrix
    # resize_mode = text_data.get("resizeMode")
    # height = text_data.get("height")
    # width = text_data.get("width")

    # styledText
    styled_text = NSKeyedUnarchiver(base64.b64decode(text_data["attributedText"]))
    string = styled_text["NSString"]
    styles = t.decode_old_text(styled_text)
    styled_text_list = read_styled_text(styles, text_element)

    return VNTextElement(
        string=string,
        transform=transform,
        styledText=styled_text_list,
        textProperty=text_property,  # TODO Illegal format
        **base_element,
    )


def read_styled_text(
    styles: List[Dict], global_style: styledElementData
) -> List[singleStyledText]:
    styled_text_list: List[singleStyledText] = []
    # inkex.utils.debug(styles)
    for style in styles:
        color = None
        # color
        if style.get("fillColor") is not None:
            color = VNColor(style["fillColor"])
        # if there's no styles in text data, use global style
        elif global_style.color is not None:
            color = global_style.color

        # stroke # TODO text strokestyle(fix reader/text.py)
        # if style.get("strokeStyle") is not None:
        #    stroke = style["strokeStyle"]
        #    stroke = pathStrokeStyle(stroke_style, stroke)

        styled_text = singleStyledText(
            length=style.get("length"),
            fontName=style.get("fontName"),
            fontSize=style.get("fontSize"),
            alignment=style.get("alignment"),
            kerning=style.get("kerning"),
            lineHeight=style.get("lineHeight"),
            fillColor=color,
            fillGradient=None,  # TODO gradient applies globally
            strokeStyle=None,
            strikethrough=style.get("strikethrough"),
            underline=style.get("underline"),
        )
        styled_text_list.append(styled_text)

    return styled_text_list


def read_vn_stroke(stylable: Dict) -> pathStrokeStyle:
    """Reads stroke style and returns as class."""
    stroke_style = stylable.get("strokeStyle")
    if stroke_style is not None:
        stroke_style["basicStrokeStyle"] = {
            "cap": stroke_style["cap"],
            "dashPattern": stroke_style["dashPattern"],
            "join": stroke_style["join"],
            "position": stroke_style["position"],
        }
        return pathStrokeStyle(
            basicStrokeStyle=basicStrokeStyle(**stroke_style["basicStrokeStyle"]),
            color=VNColor(color_dict=stroke_style["color"]),
            width=stroke_style["width"],
        )


def read_vn_fill(
    stylable: Dict, single_style: Dict = None
) -> Union[VNGradient, VNColor, None]:
    """Reads fill data and returns as class."""
    # fill
    fill_data = stylable.get("fill")
    fill_color = stylable.get("fillColor")
    fill_gradient = stylable.get("fillGradient")

    # Vectornator 4.13.2, format 13
    if fill_data is not None:
        color: Dict = fill_data.get("color", {}).get("_0")
        gradient: Dict = fill_data.get("gradient", {}).get("_0")
        if gradient is not None:
            return VNGradient(
                fill_transform=stylable["fillTransform"],
                transform_matrix=stylable["fillTransform"]["transform"],
                stops=gradient["stops"],
                typeRawValue=gradient["typeRawValue"],
            )
        elif color is not None:
            return VNColor(color_dict=color)

    # singleStyles (Vectornator 4.13.6, format 19)
    if single_style is not None and single_style.get("fill") is not None:
        fill_data = single_style["fill"]
        color: Dict = fill_data.get("color", {}).get("_0")
        gradient: Dict = fill_data.get("gradient", {}).get("_0")
        if gradient is not None:
            return VNGradient(
                fill_transform=stylable["fillTransform"],
                transform_matrix=stylable["fillTransform"]["transform"],
                stops=gradient["stops"],
                typeRawValue=gradient["typeRawValue"],
            )
        elif color is not None:
            return VNColor(color_dict=color)

    # Vectornator 4.10.4, format 8
    elif fill_color is not None or fill_gradient is not None:
        if fill_gradient is not None:
            return VNGradient(
                fill_transform=stylable["fillTransform"],
                transform_matrix=stylable["fillTransform"]["transform"],
                stops=fill_gradient["stops"],
                typeRawValue=fill_gradient["typeRawValue"],
            )
        elif fill_color is not None:
            return VNColor(color_dict=fill_color)
